module CharBuilder(buildCharacter, compileChar) where

import Types
import Plan
import AllSkills
import Compiler

import Data.Maybe
import qualified Data.Set as S
import qualified Data.Map as M
import Data.Monoid

buildCharacter :: Maybe String -> (Int, Int, Int) -> [(Maybe String, SkillPattern)] -> [Skill] -> [Plan] -> Character
buildCharacter mname (vit, refl, int) pats skills plans = Ch
  { chName = mname
  , chVit = vit
  , chReflex = refl
  , chInt = int
  , chSkillPats = map snd pats
  , chSkillPatNames = buildNames $ map fst pats
  , chSkills = skills
  , chPlans = plans
  , chPlanHints = map (genHint int) $ plans
  }

compileChar :: Character -> ([String], Character)
compileChar ch = (concat msgs, ch')
  where
    ch' = ch{ chPlans = plans' }
    (plans', msgs) = unzip $ map compilePlan $ chPlans ch
    compilePlan (Plan c) = (Plan c', msg)
      where
        (msg, c') = compileCode c
    compileCode (Code e _) = (msg, Code e (Just c))
      where
        (msg, c) = compile e

buildNames :: [Maybe String] -> [String]
buildNames mns = compl mns $ filter (`notElem` catMaybes mns) defaultNames
  where
    compl [] _ = []
    compl (Nothing:xs) (a:as) = a : compl xs as
    compl (Just x:xs) as = x : compl xs as
    defaultNames = map (:[]) ['A'..'Z'] ++ map (('Z':) . show) [1..]

genHint :: Int -> Plan -> PlanHint
genHint _int (Plan (Code ac _)) = hintOnExpr ac
  where
    hintOnExpr :: Expr -> PlanHint
    hintOnExpr e = case e of
      ConstE (FunctionV (Fun name _)) -> hintOnPrim name
      ConstE _ -> mempty
      AbsE _ body -> hintOnExpr body
      AppE fun@(RefE name) args@(RefE skname : _) -> case M.lookup skname skillEnv of
        Just sk
          | elem name priorityPrims -> mempty{ phSkillPriority = True }
          | elem name targetPrims -> mempty{ phSkillTarget = S.singleton sk }
        _ -> visit $ fun : args
      AppE fun args -> visit $ fun : args
      IfE cond tb fb -> visit [cond, tb, fb]
      LetE _ bi bo -> visit [bi, bo]
      LetNameE _ bi bo -> visit [bi, bo]
      TimeShiftE a b -> visit [a, b]
      RefE name -> hintOnPrim name
      where
        visit = mconcat . map hintOnExpr

    hintOnPrim :: String -> PlanHint
    hintOnPrim name = case name of
      "switch_to" -> mempty{ phSkillPattern = True }
      "hold" -> mempty{ phHold = True }
      "modify_plan" -> mempty{ phModify = True }
      _ | elem name priorityPrims -> mempty{ phSkillPriority = True }
        | elem name targetPrims -> mempty{ phSkillTarget = allSkillSet }
        | otherwise -> mempty

allSkillSet :: S.Set Skill
allSkillSet = S.fromList allSkills

skillEnv :: M.Map String Skill
skillEnv = M.fromList $ map f allSkills
  where
    f sk = (skName sk, sk)

priorityPrims :: [String]
priorityPrims = ["dont_use", "prioritize", "prioritize_bottom", "prioritize_above", "prioritize_below", "reserve", "set_reaction_count", "use"]

targetPrims :: [String]
targetPrims = ["target", "target_with_priorities"]
