module CommonOperations where

import Game
import Run
import GameUtils

import Control.Monad
import Data.List
import Data.Maybe
import qualified Data.Set as S

getSDs :: CharDescr -> Game [SlotDescr]
getSDs cd = do
  sss <- getSSs cd
  let build ss = SD cd (ssIndex ss) (ssSkill ss)
  return $ map build sss

plannedSkills :: CharDescr -> (Skill -> Bool) -> Maybe Invocation -> Maybe Attack -> Game [(SlotDescr, Bool)]
plannedSkills cd prop_hint minv matk = do
  readiness <- skillsReadiness cd
  let ok (ss, r) = r && prop_hint (ssSkill ss)
  let readies = filter ok readiness
  if null readies then return [] else do -- shortcut
    pri <- askPriority cd minv matk
    let 
      f ((ss, True), force) = Just (SD cd (ssIndex ss) $ ssSkill ss, force)
      f _ = Nothing
    let rs = mapMaybe f $ pri (ssIndex . fst) readiness
    return rs

sdReady :: SlotDescr -> Game Bool
sdReady (SD cd slot sk) = do
  af <- canAfford cd sk
  ub <- liftM ((==0) . ssBound . fromJust) $ getSS cd slot
  return $ af && ub

activateReactiveSkillsAgainstSkill :: CharDescr -> Invocation -> Game Bool
activateReactiveSkillsAgainstSkill cd inv = do
  sds <- plannedSkills cd (isJust . skReactionAgainstSkill) (Just inv) Nothing
  f sds
  where
    source = invSkill inv
    f [] = return False
    f ((sd@(SD _ _ sk), forcep):rest) =
      require (isJust $ skReactionAgainstSkill sk) $ do
        ready <- sdReady sd
        require ready $ do
          cond <- skReactionConditionAgainstSkill sk inv
          require (evalSkillCondition forcep cond) $ do
            neg <- reactiveInvokeAgainstSkill sd inv
            require neg $ do
              mesC (opponent cd) $ "の発動した" ++ show source ++ "の効果は打ち消された"
              return True
      where
        require c u = if c then u else f rest

hold :: CharDescr -> (Unique -> HeldSkill) -> Game Unique
hold cd build = do
  un <- newUnique
  hss <- getHSs cd
  let hs = build un
  setHSs cd $ hs : hss
  hsStart hs
  return un

unhold :: CharDescr -> DropReason -> Int -> Game [Delay]
unhold cd reason idx = do
  hss <- getHSs cd
  let (pre, t:post) = splitAt idx hss 
  setHSs cd $ pre ++ post
  terminateHeldSkill reason t

hsTime :: HeldSkill -> Time
hsTime = time . hsSource

terminateHeldSkill :: DropReason -> HeldSkill -> Game [Delay]
terminateHeldSkill reason t = do
  mesC cd $ "の" ++ show (invSkill $ hsSource t) ++ "の維持が解除された"
  when (hsEffective t) $ hsStop t
  return [(hsTime t, finish)]
  where
    cd = invControler $ hsSource t
    finish = do
      (now, turn) <- tick
      recordEvent $ Drop{ dropTime = now, dropTurn = turn, dropInv = hsSource t, dropReason = reason }
      hsFinish t

tick :: Game (Time, Turn)
tick = liftM2 (,) tickTime getCurrentTurn

unholdByUnique :: CharDescr -> DropReason -> Unique -> Game [Delay]
unholdByUnique cd reason un = do
  hs <- getHSs cd
  case findIndex ((==un) . hsUnique) hs of
    Just idx -> unhold cd reason idx
    Nothing -> fail "unholdByUnique: hs not found"

rollPanel :: CharDescr -> Int -> Int -> Game [Delay]
rollPanel cd m n = do
  oldk <- liftM paKind $ getPanel' cd m n
  (k, ds) <- rollPanel' cd m n oldk
  modifyPanel cd m n [SetKind k]
  return ds

rollPanelBase :: CharDescr -> Int -> Int -> Game [Delay]
rollPanelBase cd m n = do
  oldk <- liftM paKind $ getPanel' cd m n
  (k, ds) <- rollPanel' cd m n oldk
  modifyPanel cd m n [SetKind k, SetBaseKind k, SetConfuse 0]
  return ds

rollPanel' :: CharDescr -> Int -> Int -> PanelKind -> Game (PanelKind, [Delay])
rollPanel' cd m n k = do
  v <- rnd 6
  let k' = if v == 5 then k else toEnum v
  ds <- if k == k'
    then do
      mesPa cd m n $ "は変更なし"
      return []
    else do
      mesPa cd m n $ "が" ++ showKind k ++ "から" ++ showKind k' ++ "に変更"
      notifyPanelChange cd DropByRoll m n
  return (k', ds)

rollPanelTo :: CharDescr -> Int -> Int -> PanelKind -> Game [Delay]
rollPanelTo cd m n k = do
  cur <- liftM (paKind . fromJust) $ getPanel cd m n
  if (cur == k) then return [] else do
    mesPa cd m n $ "が" ++ showKind k ++ "に変更"
    modifyPanel cd m n [SetKind k, SetBaseKind k, SetConfuse 0]
    notifyPanelChange cd DropByRoll m n

notifyPanelChange :: CharDescr -> DropReason -> Int -> Int -> Game [Delay]
notifyPanelChange cd reason m n = do
  hss <- getHSs cd
  case findIndex ((elem (m, n)) . hsCosts) hss of
    Just idx -> unhold cd reason idx
    _ -> return []

invoke :: SlotDescr -> InvocationType -> Game (Invocation, Bool)
invoke (SD cd slot sk) typ = do
  nt <- tick
  costs <- mapM (allocPanel cd) $ skCosts sk
  invokeWithCosts cd slot sk typ nt costs

invokeWithCosts :: CharDescr -> Int -> Skill -> InvocationType -> (Time, Turn) -> [(Int, Int)] -> Game (Invocation, Bool)
invokeWithCosts cd slot sk typ (tm, tn) costs = do
  recordEvent inv
  neg <- activateReactiveSkillsAgainstSkill (opponent cd) inv
  return (inv, neg)
  where
    inv = Inv
      { invSlot = slot
      , invSkill = sk
      , invTime = tm
      , invTurn = tn
      , invType = typ
      , invCosts = costs
      , invControler = cd
      }

reactiveInvokeAgainstSkill :: SlotDescr -> Invocation -> Game Bool
reactiveInvokeAgainstSkill sd@(SD cd _ sk) obj = do
  mesC cd $ "は" ++ show sk ++ "を受動発動した"
  (inv, neg) <- invoke sd (ReactionAgainstSkillI obj)
  if neg
    then return False
    else case skReactionAgainstSkill sk of
      Just handler -> handler inv obj

reactiveInvoke :: SlotDescr -> Attack -> Game (AttackReaction, Invocation)
reactiveInvoke sd@(SD cd _ sk) src = do
  mesC cd $ "は" ++ show sk ++ "を受動発動した"
  (inv, neg) <- invoke sd (ReactionI src)
  if neg
    then return (nullReaction, inv)
    else case skReaction sk of
      Just reaction -> return (reaction, inv)

activeInvoke :: SlotDescr -> Game ()
activeInvoke sd@(SD cd _ sk) = do
  mesC cd $ "は" ++ show sk ++ "を発動した"
  (inv, neg) <- invoke sd ActionI
  when (not neg) $ case skEffect sk of
    Just fn -> fn inv

nullReaction :: AttackReaction
nullReaction _ n _ = return (n, False, const $ return ())

toUsed :: CharDescr -> Int -> Int -> Game ()
toUsed cd m n = do
  Just pa <- getPanel cd m n
  when (not $ paUsed pa) $ do
    replicateM_ (paPoisoned pa) $ damage cd ByPoison 1
    modifyPanel cd m n [SetUsed True]

allocPanel :: CharDescr -> PanelKind -> Game (Int, Int)
allocPanel cd pk = do
  csp <- getCurrentSP' cd
  sps <- getPanels cd
  p <- choose $ filter usable sps
  let m = paIndex p
  toUsed cd csp m
  return (csp, m)
  where
    usable pa = paUsable pa && paKind pa == pk

genDmg :: CharDescr -> DamageReason -> Int -> Game Damage
genDmg cd reason n = do
  (now, turn) <- tick
  let
    dmg = Dmg
      { dmgTime = now
      , dmgTurn = turn
      , dmgReason = reason
      , dmgAmount = n
      , dmgSufferer = cd
      }
  recordEvent dmg
  return dmg

damage :: CharDescr -> DamageReason -> Int -> Game Damage
damage cd reason n = do
    sd <- isSuddenDeath
    when sd $ do
      mes $ "サドンデス成立"
      lose cd
    mesC cd $ "に" ++ show n ++ "点のダメージ"
    chp <- getHp cd
    let newhp = chp - n
    setHp cd newhp
    sara <- getSarasouju cd
    when (newhp <= 0 && sara == 0) $ lose cd
    replicateM_ sara $ decrVit cd 1
    genDmg cd reason n

cureConfuse :: CharDescr -> Int -> Int -> Game [Delay]
cureConfuse cd m n = do
  mesPa cd m n $ "のConfuseが回復した"
  pa <- getPanel' cd m n
  let conf = paConfused pa - 1
  (ops, changed) <- if conf > 0 || paKind pa == paBaseKind pa
    then return ([SetConfuse conf], False) 
    else do
      let k = paBaseKind pa
      mesPa cd m n $ "のConfuseがすべて消えたため、このパネルは" ++ showKind k ++ "に戻った"
      return ([SetConfuse conf, SetKind k], True)
  modifyPanel cd m n ops
  if changed 
    then notifyPanelChange cd DropByRoll m n
    else return []

curePanelBind :: CharDescr -> Int -> Int -> Game ()
curePanelBind cd m n = do
  mesPa cd m n $ "のBindが回復した"
  modifyPanel cd m n [AddBind (-1)]

cureSkillBind :: CharDescr -> Int -> Game ()
cureSkillBind cd i = do
  mesC cd $ "のスキル" ++ show i ++ "のBindが回復した"
  addSkillBind cd i (-1)

curePlanBind :: CharDescr -> Int -> Game ()
curePlanBind cd i = do
  mesC cd $ "のプラン" ++ show i ++ "のBindが回復した"
  addPlanBind cd i (-1)

loseSlot :: CharDescr -> Int -> Game [Delay]
loseSlot cd n = do
  mesC cd $ "はスキル" ++ show n ++ "を失った"
  deleteSlot cd n
  hss <- getHSs cd
  liftM concat $ mapM (unholdByUnique cd DropByLostSkill . hsUnique) $ filter ((==n) . invSlot . hsSource) hss

losePlan :: CharDescr -> Int -> Game ()
losePlan cd n = do
  mesC cd $ "はプラン" ++ show n ++ "を失った"
  deletePlan cd n

losePanel :: CharDescr -> Int -> Int -> Game [Delay]
losePanel cd m n = do
  mesC cd $ "はスキルパターン" ++ show m ++ "のパネル" ++ show n ++ "を失った"
  deletePanel cd m n
  notifyPanelChange cd DropByLostPanel m n
