module Compiler(compile) where

import Plan
import ExprUtils

import qualified Data.Map as M
import qualified Data.Set as S
import qualified Data.Sequence as Q
import Data.Sequence(Seq, (<|), (><))
import Control.Monad.State
import GlobalEnv
import Control.Monad.Fix

type REnv = Q.Seq Value -- runtime environment
type Compiler = State [String]
type LEnv = DMap Name

data DMap k = DMap !Int !(M.Map k Int)

emptyD :: (Ord k) => DMap k
emptyD = DMap 0 M.empty

insertD :: (Ord k) => k -> DMap k -> DMap k
insertD k (DMap sz m) = DMap (sz+1) (M.insert k sz m)

lookupD :: (Ord k) => k -> DMap k -> Maybe Int
lookupD k (DMap sz m) = do
  v <- M.lookup k m
  return $ sz - v - 1

runCompiler :: Compiler a -> ([String], a)
runCompiler action = (msgs, r)
  where
    (r, msgs) = runState action []

compile :: Expr -> ([String], Eval Value)
compile exp = runCompiler $ do
  code <- comp S.empty globalEnv emptyD exp
  return $ code Q.empty

comp :: S.Set Name -> Env -> LEnv -> Expr -> Compiler (REnv -> Eval Value)
comp recs env lenv exp = case exp of
  ConstE v -> return $ \re -> return v
  AbsE vars body -> do
    let topenv = foldr insertD lenv vars
    code <- comp S.empty env topenv body
    let arity = length vars
    return $ \re -> {-# SCC "FunctionCreation" #-} do
      tm <- getNow
      u <- genUniqueName "compf"
      return $ FunctionV $ Fun u $ \args -> {-# SCC "FunctionEntrance" #-}
        if length args == arity
          then withETime tm $ code (foldr (<|) re args)
          else err $ "wrong number of arguments passed to function " ++ show vars ++ " => " ++ show body
  AppE (ConstE (FunctionV fcn)) args -> directApp fcn args
  AppE fn@(RefE fname) args -> case lookupD fname lenv of
    Just _ -> app fn args -- only optimize for global functions
    Nothing -> case M.lookup fname env of
      Just (Value (FunctionV fcn)) -> directApp fcn args
      _ -> app fn args
  AppE fn args -> app fn args
  IfE cond tb fb -> do
    cond_c <- rec cond
    tb_c <- rec tb
    fb_c <- rec fb
    return $ \re -> {-# SCC "If" #-} do
      c <- cond_c re
      case c of
        BoolV True -> tb_c re
        BoolV False -> fb_c re
        _ -> err $ "condition value is not a boolean: " ++ show c
  LetE var bi bo -> do
    let lenv' = insertD var lenv
    bi_code <- comp (S.insert var recs) env lenv' bi
    bo_code <- comp (S.delete var recs) env lenv' bo
    return $ \re -> {-# SCC "Let" #-} do
      (re', _) <- mfix $ \ ~(_, v) -> do
        let re' = v <| re
        rv <- bi_code re'
        return (re', rv)
      bo_code re'
  LetNameE var bi bo -> comp recs (M.insert var (Thunk bi) env) lenv bo
  TimeShiftE tm body -> do
    tm_c <- rec tm
    body_c <- rec body
    return $ \re -> {-# SCC "TimeShift" #-} do
      t <- tm_c re
      case t of
        TimeV s -> withETime s $ body_c re
        _ -> err $ "first operand of # is not a time: " ++ show t
  RefE var -> case lookupD var lenv of
    Just n
      | S.member var recs -> do
          warn $ "variable " ++ var ++ " is used directly in its definition"
          return $ \re -> err $ "circular definition: " ++ var
      | otherwise -> return $ \re -> {-# SCC "Ref" #-} return $ Q.index re n
    Nothing -> case M.lookup var env of
      Just (Value v) -> return $ \re -> return v
      Just (Thunk e) -> rec (LetE var e (RefE var))
      Nothing -> do
        warn $ "unbound variable: " ++ showStringLit var
        return $ \re -> err $ "undefined variable: " ++ showStringLit var
  where
    rec = comp recs env lenv
    app fn args = do
      fn_c <- rec fn
      arg_cs <- mapM rec args
      return $ case arg_cs of
        [arg_c] -> \re -> {-# SCC "UnaryApplication" #-} do
          fn <- fn_c re
          val <- arg_c re
          apply1 fn val
        [arg_c0, arg_c1] -> \re -> {-# SCC "BinaryApplication" #-} do
          fn <- fn_c re
          val0 <- arg_c0 re
          val1 <- arg_c1 re
          apply2 fn val0 val1
        _ -> \re -> {-# SSC "GeneralApplication" #-} do
          let exec f = f re
          fn <- fn_c re
          values <- mapM exec arg_cs
          apply fn values
    directApp fcn args = do
      arg_cs <- mapM rec args
      case fcn of
        Fun1 _ f -> case arg_cs of
          [arg_c] -> return $ \re -> {-# SCC "UnaryDirectApplication" #-} do
            v <- arg_c re
            f v
          _ -> warnArgMismatch 1 fcn args
        Fun2 _ f -> case arg_cs of
          [arg_c0, arg_c1] -> return $ \re -> {-# SCC "BinaryDirectApplication" #-} do
            v0 <- arg_c0 re
            v1 <- arg_c1 re
            f v0 v1
          _ -> warnArgMismatch 2 fcn args
        Fun _ f -> case arg_cs of
          [] -> return $ \re -> {-# SCC "ThunkApplication" #-} f []
          _ -> return $ \re -> {-# SCC "GeneralDirectApplication" #-} do
            let exec v = v re
            vs <- mapM exec arg_cs
            f vs
    warnArgMismatch n fcn args = do
      warn $ "function " ++ show (FunctionV fcn) ++ " requires " ++ show n ++ " arguments but passed " ++ show args
      return $ \_ -> argumentNumberError n fcn args

warn :: String -> Compiler ()
warn msg = modify (++[msg])
