{-# OPTIONS_GHC -fglasgow-exts #-}
module ExprUtils where

import Game
import GameUtils

import Control.Exception
import Data.Dynamic
import Control.Monad
import Control.Monad.Reader
import qualified Data.Set as S
import qualified Data.Map as M
import Plan

always :: Expr
always = ConstE $ BoolV True

fromSkillTarget :: SkillTarget -> Value
fromSkillTarget st = case st of
  PanelTarget cd m n -> PanelV cd m n
  SkillTarget cd n -> SkillSlotV cd n
  PlanTarget cd n -> PlanV cd n

toInteg :: (MonadIO m) => String -> Value -> m Integer
toInteg _ (IntegerV v) = return v
toInteg loc v = typeError loc v "integer"

integerToInt :: Integer -> Maybe Int
integerToInt i
  | minb <= i && i <= maxb = Just $ fromIntegral i
  | otherwise = Nothing
  where
    minb = fromIntegral (minBound :: Int)
    maxb = fromIntegral (maxBound :: Int)

toInt :: (MonadIO m) => String -> Value -> m Int
toInt loc v = do
  i <- toInteg loc v
  case integerToInt i of
    Just v -> return v
    Nothing -> err $ loc ++ ": integral value out of bounds: " ++ show i

toBool :: (MonadIO m) => String -> Value -> m Bool
toBool _ (BoolV v) = return v
toBool loc v = typeError loc v "boolean"

toRch :: (MonadIO m) => String -> Value -> m RCh
toRch _ (RChV v) = return v
toRch loc v = typeError loc v "relative character description"

toSkill :: (MonadIO m) => String -> Value -> m Skill
toSkill _ (SkillV v) = return v
toSkill loc v = typeError loc v "skill"

toSkillSlot :: (MonadIO m) => String -> Value -> m (CharDescr, Int)
toSkillSlot _ (SkillSlotV cd v) = return (cd, v)
toSkillSlot loc v = typeError loc v "skill slot"

toPanel :: (MonadIO m) => String -> Value -> m (CharDescr, Int, Int)
toPanel _ (PanelV cd m n) = return (cd, m, n)
toPanel loc v = typeError loc v "panel"

toList :: (MonadIO m) => String -> Value -> m [Value]
toList _ (ListV es) = return es
toList loc v = typeError loc v "list"

toPlan :: (MonadIO m) => String -> Value -> m (CharDescr, Int)
toPlan _ (PlanV cd n) = return (cd, n)
toPlan loc v = typeError loc v "plan"

toPanelKind :: (MonadIO m) => String -> Value -> m PanelKind
toPanelKind _ (PanelKindV k) = return k
toPanelKind loc v = typeError loc v "panel kind"

toAction :: (MonadIO m) => String -> Value -> m PlanAction
toAction _ (ActionV a) = return a
toAction loc v = typeError loc v "command"

toSet :: (MonadIO m) => String -> Value -> m (S.Set Value)
toSet _ (SetV a) = return a
toSet loc v = typeError loc v "set"

toSet' :: (MonadIO m) => String -> Value -> m (S.Set Value)
toSet' _ (SetV a) = return a
toSet' _ (ListV l) = return $ S.fromList l
toSet' loc v = typeError loc v "set or list"

toList' :: (MonadIO m) => String -> Value -> m [Value]
toList' _ (SetV a) = return $ S.toList a
toList' _ (ListV l) = return l
toList' loc v = typeError loc v "set or list"

toAgg :: (MonadIO m) => String -> Value -> m (Either (S.Set Value) [Value])
toAgg _ (SetV a) = return $ Left a
toAgg _ (ListV a ) = return $ Right a
toAgg loc v = typeError loc v "set or list"

toInv :: (MonadIO m) => String -> Value -> m Invocation
toInv _ (InvocationV inv) = return inv
toInv loc v = typeError loc v "invocation"

toTurn :: (MonadIO m) => String -> Value -> m Turn
toTurn _ (TurnV t) = return t
toTurn loc v = typeError loc v "turn"

toAttack :: (MonadIO m) => String -> Value -> m Attack
toAttack _ (AttackV atk) = return atk
toAttack loc v = typeError loc v "attack"

toIntf :: (MonadIO m) => String -> Value -> m Interference
toIntf _ (InterferenceV int) = return int
toIntf loc v = typeError loc v "block"

toString :: (MonadIO m) => String -> Value -> m String
toString _ (StringV str) = return str
toString loc v = typeError loc v "string"

toDmg :: (MonadIO m) => String -> Value -> m Damage
toDmg _ (DamageV dmg) = return dmg
toDmg loc v = typeError loc v "damage"

toSkillPattern :: (MonadIO m) => String -> Value -> m (CharDescr, Int)
toSkillPattern _ (SkillPatternV cd v) = return (cd, v)
toSkillPattern loc v = typeError loc v "skill pattern"

toCont :: (MonadIO m) => String -> Value -> m Continue
toCont _ (ContinuationV cont) = return cont
toCont loc v = typeError loc v "continuation"

toDrop :: (MonadIO m) => String -> Value -> m Drop
toDrop _ (DropV drp) = return drp
toDrop loc v = typeError loc v "drop"

toRandom :: (MonadIO m) => String -> Value -> m RndGen
toRandom _ (RandomV (Random gen)) = return gen
toRandom loc v = typeError loc v "random"

toFunction :: (MonadIO m) => String -> Value -> m Function
toFunction _ (FunctionV fun) = return fun
toFunction loc v = typeError loc v "function"

typeError :: (MonadIO m) => String -> Value -> String -> m a
typeError loc v typ = err $ loc ++ ": " ++ typ ++ " expected but got " ++ show v

newtype EvalError = EvalError String
  deriving (Typeable)

err :: (MonadIO m) => String -> m a
err s = liftIO $ throwIO $ DynException $ toDyn $ EvalError s

runEval :: CharDescr -> Eval a -> Game (Either String a)
runEval cd action = do
  gen <- newRndGen
  now <- tickTime
  runEvalWithContext cd action EC{ ecCurrentInv = Nothing, ecCurrentAtk = Nothing, ecRndGen = gen, ecNow = now }

runEvalWithContext :: CharDescr -> Eval a -> EvalContext -> Game (Either String a)
runEvalWithContext cd action ec = do
  gi <- ask
  liftIO $ f $ runGame (unEval action cd ec) gi
  where
    f a = catchJust evalErrors (liftM Right a) (return . Left)

evalErrors :: Exception -> Maybe String
evalErrors exn = do
  v <- dynExceptions exn
  EvalError s <- fromDynamic v
  return s

binIntPrim :: String -> (Integer -> Integer -> Eval Value) -> Expr
binIntPrim name fn = binPrimE name $ \x y -> do
  xi <- liftGame $ toInteg name x
  yi <- liftGame $ toInteg name y
  fn xi yi

unPrimE :: String -> (Value -> Eval Value) -> Expr
unPrimE name fn = ConstE $ unPrim name fn

unPrim :: String -> (Value -> Eval Value) -> Value
unPrim name fn = FunctionV (Fun1 name fn)

binPrimE :: String -> (Value -> Value -> Eval Value) -> Expr
binPrimE name fn = ConstE $ binPrim name fn

binPrim :: String -> (Value -> Value -> Eval Value) -> Value
binPrim name fn = FunctionV (Fun2 name fn)

terPrim :: String -> (Value -> Value -> Value -> Eval Value) -> Value
terPrim name fn = FunctionV (Fun name f)
  where
    f [x, y, z] = fn x y z
    f _ = err $ "wrong number of arguments passed to ternary function " ++ name

unPrim' :: String -> (CharDescr -> Value -> Game Value) -> Value
unPrim' name fn = unPrim name $ \x -> do
  cd <- getSelf
  liftGame $ fn cd x

unPrim0 :: String -> (Value -> Game Value) -> Value
unPrim0 name fn = unPrim' name fn'
  where
    fn' _ = fn

binPrim0 :: String -> (Value -> Value -> Game Value) -> Value
binPrim0 name fn = binPrim' name fn'
  where
    fn' _ = fn

binPrim' :: String -> (CharDescr -> Value -> Value -> Game Value) -> Value
binPrim' name fn = binPrim name $ \x y -> do
  cd <- getSelf
  liftGame $ fn cd x y

terPrim' :: String -> (CharDescr -> Value -> Value -> Value -> Game Value) -> Value
terPrim' name fn = terPrim name $ \x y z -> do
  cd <- getSelf
  liftGame $ fn cd x y z

unIntPrim' :: String -> (CharDescr -> Int -> Game Value) -> Value
unIntPrim' name fn = unPrim' name $ \cd x -> fn cd =<< toInt name x

valueCompare :: String -> Value -> Value -> Eval Ordering
valueCompare name x y = case (x, y) of
  (IntegerV i, IntegerV j) -> return $ compare i j
  (StringV i, StringV j) -> return $ compare i j
  (TimeV i, TimeV j) -> return $ compare i j
  (ListV i, ListV j) -> onList i j
  _ -> err $ name ++ ": bad comparison between " ++ show x ++ " and " ++ show y
  where
    onList [] [] = return EQ
    onList [] _ = return LT
    onList _ [] = return GT
    onList (x:xs) (y:ys) = do
      r <- valueCompare name x y
      case r of
        EQ -> onList xs ys
        _ -> return r

equal :: Value -> Value -> Eval Bool
equal x@(FunctionV _) y = err $ "cannot test equality on functions: " ++ show x ++ ", " ++ show y
equal x y = return $ x == y

rel :: CharDescr -> RCh -> CharDescr
rel cd Self = cd
rel cd Opp = opponent cd

withETime ::  ETime -> Eval a -> Eval a
withETime t x = do
  snapshots <- liftGame getSnapshots
  x' <- unliftEval $ withNow t x
  liftGame $ withGS (snapshots M.! t) x'

applyFun :: Function -> [Value] -> Eval Value
applyFun fcn args = case fcn of
  Fun _ f -> f args
  Fun1 _ f -> case args of
    [x] -> f x
    _ -> argumentNumberError 1 fcn args
  Fun2 _ f -> case args of
    [x, y] -> f x y
    _ -> argumentNumberError 2 fcn args

applyFun1 :: Function -> Value -> Eval Value
applyFun1 fcn arg = case fcn of
  Fun _ f -> f [arg]
  Fun1 _ f -> f arg
  Fun2 _ _ -> argumentNumberError 2 fcn [arg]

applyFun2 :: Function -> Value -> Value -> Eval Value
applyFun2 fcn arg0 arg1 = case fcn of
  Fun _ f -> f [arg0, arg1]
  Fun1 _ _ -> argumentNumberError 1 fcn [arg0, arg1]
  Fun2 _ f -> f arg0 arg1

checkFunction :: Value -> Eval Function
checkFunction = toFunction "application"

apply :: Value -> [Value] -> Eval Value
apply fn args = do
  fcn <- checkFunction fn
  applyFun fcn args

apply1 :: Value -> Value -> Eval Value
apply1 fn arg = do
  fcn <- checkFunction fn
  applyFun1 fcn arg

apply2 :: Value -> Value -> Value -> Eval Value
apply2 fn arg arg1 = do
  fcn <- checkFunction fn
  applyFun2 fcn arg arg1

argumentNumberError n fcn args =
  err $ "wrong number of arguments passed to " ++ show (FunctionV fcn) ++ ": required " ++ show n ++ ", but got " ++ show args

genUniqueName :: String -> Eval Name
genUniqueName prefix = do
  u <- liftGame newUnique
  return $ ":" ++ prefix ++ show u

intV :: Int -> Value
intV = IntegerV . fromIntegral

rndgenV :: RndGen -> Value
rndgenV = RandomV . Random

newFun1 :: String -> (Value -> Eval Value) -> Eval Value
newFun1 base f = do
  name <- genUniqueName base
  return $ FunctionV (Fun1 name f)

planExpr :: Plan -> Expr
planExpr = codeExpr . planCode

planCode :: Plan -> Code
planCode (Plan c) = c
