{-# OPTIONS_GHC -fglasgow-exts #-}
module Game
  ( module Types
  -- the Game monad
  , Game
  , runGame
  , GameInfo
  , RndGen
  , GameFinishedException(..)
  , GameResult(..)
  , newGameInfo
  -- primitive operations
  , isSuddenDeath
  , gameFinally
  , newUnique
  , mes
  , tickTime
  , getNextTime
  , isOutputPresent
  -- random number generation
  , rnd
  , getRndGen
  , newRndGen
  -- other types
  , Skill(..)
  , SkillCondition(..)
  , SkillAttribute(..)
  , AttackReaction
  , SkillReaction
  , SlotDescr(..)
  , Turn(..)
  , Time
  , initialTime
  , Unique
  , initialUnique
  -- events
  , Invocation(..)
  , InvocationType(..)
  , Attack(..)
  , AttackResult(..)
  , Interference(..)
  , InterferenceType(..)
  , Damage(..)
  , DamageReason(..)
  , Continue(..)
  , Drop(..)
  , DropReason(..)
  , Event(time, turn)
  , Cal
  -- accessors
  -- * skill patterns
  , getSPSs
  , getSPS
  , getSPSByName
  , getPanels
  , getPanel
  , getPanel'
  , getCurrentSP
  , withCurrentSP
  , setCurrentSP
  , deleteCurrentSP
  , modifyPanel
  , modifyCurrentPanels
  , deleteSP
  , undeleteSP
  , deletePanel
  , undeletePanel
  , PanelOperation(..)
  , Panel(..)
  , SkillPatState(..)
  -- * skills
  , getSSs
  , getSS
  , addSkillBind
  , addSkillWeaken
  , deleteSlot
  , undeleteSlot
  , SkillState(..)
  -- * plans
  , addPlanBind
  , deletePlan
  , undeletePlan
  , getPresentPlans
  , getPlanState
  , PlanState(..)
  -- * game history
  , getCalendar
  , getTurns
  , getSnapshots
  , recordEvent
  , recordTurn
  -- * held skills
  , getHSs
  , setHSs
  , modifyHS
  , HeldSkill(..)
  -- * other character-wise state
  , getHp
  , setHp
  , getVit
  , setVit
  , getBaseVit
  , setBaseVit
  , getRefl
  , setRefl
  , getInt
  , setInt
  , getSarasouju
  , modifySarasouju
  , getAngelFire
  , modifyAngelFire
  -- * game state object
  , GameState
  , getGS
  , withGS
  , getGSFromGI
  -- * turns
  , getCurrentTurn
  , setCurrentTurn
  -- * character info
  , char
  -- * paradise lost
  , getParadiseLost
  , addParadiseLost
  -- * reaction attach
  , registerReaction
  , getReactions
  -- utilities
  , vitToMaxHp
  , sortOn
  ) where

import Types

import qualified Data.Set as S
import qualified Data.Map as M
import Data.Maybe
import Data.List
import Data.Typeable
import Control.Monad.Reader
import Data.IORef
import System.Random
import Control.Exception

newtype Game a = Game (GameInfo -> IO a)

instance Monad Game where
  return a = Game $ \gi -> gi `seq` return a
  Game g >>= f = Game $ \gi -> gi `seq` do
    v <- g gi
    case f v of
      Game sth -> sth gi

instance MonadReader GameInfo Game where
  ask = Game $ \gi -> gi `seq` return gi
  local f (Game g) = Game $ \gi -> gi `seq` let gi' = f gi in gi' `seq` g gi'

instance MonadFix Game where
  mfix f = Game $ \gi -> mfix (\v -> runGame (f v) gi)

instance MonadIO Game where
  liftIO ioa = Game $ \gi -> gi `seq` ioa

runGame :: Game a -> GameInfo -> IO a
runGame (Game f) gi = f gi

data GameInfo = GI
  { giChar0 :: {-# UNPACK #-} !Character
  , giChar1 :: {-# UNPACK #-} !Character
  , giState :: {-# UNPACK #-} !(IORef GameState)
  , giHistory :: {-# UNPACK #-} !(IORef GameHistory)
  , giUniqueSource :: {-# UNPACK #-} !(IORef Unique)
  , giIsSuddenDeath :: {-# UNPACK #-} !Bool
  , giOutput :: {-# UNPACK #-} !(Maybe (String -> IO ()))
  , giTickHook :: !(Game ())
  , giRndGen :: !(IORef RndGen)
  }

type RndGen = StdGen

data GameState = GS
  { gsCurrentTurn :: !(Maybe Turn)
  , gsChar0 :: !CharState
  , gsChar1 :: !CharState
  , gsTimeSource :: !Time
  , gsParadiseLost :: !Int
  , gsReactions :: !(M.Map Invocation [ReactionInfo])
  }
  deriving (Show, Eq)

data ReactionInfo = RI AttackReaction Invocation

instance Show ReactionInfo where
  show (RI _ inv) = show inv

instance Eq ReactionInfo where
  RI _ inv0 == RI _ inv1 = inv0 == inv1

data Turn = Turn
  { turnNumber :: !Int
  , turnTime :: !Time
  , turnFirstToMove :: !CharDescr
  }
  deriving (Show, Eq, Ord)

newtype Time = Time Int
  deriving (Eq, Ord, Enum)

instance Show Time where
  show (Time n) = show n

initialTime :: Time
initialTime = Time 0

newtype Unique = Unique Int
  deriving (Eq, Ord, Enum)

instance Show Unique where
  show (Unique n) = "u" ++ show n

initialUnique :: Unique
initialUnique = Unique 0

data CharState = CS
  { csHp :: !Int
  , csVit :: !Int
  , csBaseVit :: !Int
  , csRefl :: !Int
  , csInt :: !Int
  , csSkillPats :: ![SkillPatState]
  , csCurrentSP :: !(Maybe Int)
  , csSkills :: ![SkillState]
  , csPlans :: ![PlanState]
  , csSkillsHeld :: ![HeldSkill]
  , csSarasouju :: !Int
  , csAngelFire :: !Int
  }
  deriving (Show, Eq)

data SkillPatState = SPS
  { spsPanels :: ![Panel]
  , spsName :: !String
  , spsIndex :: !Int
  }
  deriving (Show, Eq)

data Panel = Pa
  { paPoisoned :: !Int
  , paConfused :: !Int
  , paBound :: !Int
  , paKind :: !PanelKind
  , paUsed :: !Bool
  , paEnsnared :: !Bool
  , paBaseKind :: !PanelKind -- affected by rolling, but not by Confuse
  , paIndex :: !Int
  , paOriginalKind :: !PanelKind
  }
  deriving (Show, Eq)

data SkillState = SS
  { ssBound :: !Int
  , ssWeakened :: !Int
  , ssSkill :: !Skill
  , ssIndex :: !Int
  }
  deriving (Show, Eq)

data PlanState = PS
  { psBound :: !Int
  , psPlan :: !Plan
  , psHint :: !PlanHint
  , psIndex :: !Int
  }
  deriving (Show, Eq)

data HeldSkill = HS
  { hsSource :: Invocation
  , hsFinish :: Game ()
  , hsStart :: Game ()
  , hsStop :: Game ()
  , hsReaction :: Int -> Attack -> Game (Int, Interference -> Game ())
  , hsInterfereCondition :: Attack -> Bool
  , hsEffective :: Bool
  , hsUnique :: Unique
  }

instance Show HeldSkill where
  showsPrec n = showsPrec n . invSkill . hsSource

instance Eq HeldSkill where
  a == b = hsSource a == hsSource b

data Invocation = Inv
  { invTime :: !Time
  , invTurn :: !Turn
  , invSlot :: !Int
  , invSkill :: !Skill
  , invType :: !InvocationType
  , invCosts :: ![(Int, Int)]
  , invControler :: !CharDescr
  }
  deriving (Show, Eq, Ord)
data InvocationType = ActionI | ReactionI Attack | ReactionAgainstSkillI Invocation | ReleaseI Invocation
  deriving (Show, Eq, Ord)

data Attack = Atk
  { atkTime :: !Time
  , atkTurn :: !Turn
  , atkInv :: !Invocation
  , atkCause :: !(Maybe Interference)
  , atkClaimedDmg :: !Int
  , atkTarget :: !CharDescr
  }
  deriving (Show, Eq, Ord)

data AttackResult = AR
  { arAttack :: !Attack
  , arDamage :: !(Maybe Damage)
  , arNegated :: !Bool
  }
  deriving (Show, Eq, Ord)

data Interference = Int
  { intTime :: !Time
  , intTurn :: !Turn
  , intInput :: !Int
  , intOutput :: !Int
  , intNegated :: !Bool
  , intInv :: !Invocation
  , intType :: !InterferenceType
  , intTarget :: !Attack
  }
  deriving (Show, Eq, Ord)

data InterferenceType = InvokedI | HeldI
  deriving (Show, Eq, Ord)

data Damage = Dmg
  { dmgTime :: !Time
  , dmgTurn :: !Turn
  , dmgReason :: !DamageReason
  , dmgAmount :: !Int
  , dmgSufferer :: !CharDescr
  }
  deriving (Show, Eq, Ord)

data DamageReason = ByAttack Attack | ByPoison
  deriving (Show, Eq, Ord)

-- continuation and drop of a held skill

data Continue = Cont
  { contTime :: !Time
  , contTurn :: !Turn
  , contInv :: !Invocation
  }
  deriving (Show, Eq, Ord)

data Drop = Drop
  { dropTime :: !Time
  , dropTurn :: !Turn
  , dropInv :: !Invocation
  , dropReason :: !DropReason
  }
  deriving (Show, Eq, Ord)

data DropReason = DropByWill | DropByRoll | DropByNegation | DropByLostSkill | DropByLostPanel
  deriving (Show, Eq, Ord)

class Event a where
  time :: a -> Time
  turn :: a -> Turn
  getCal :: GameHistory -> Cal a
  setCal :: Cal a -> GameHistory -> GameHistory

instance Event Invocation where
  time = invTime
  turn = invTurn
  getCal = ghInvs
  setCal x gh = gh{ ghInvs = x }

instance Event Attack where
  time = atkTime
  turn = atkTurn
  getCal = ghAttacks
  setCal x gh = gh{ ghAttacks = x }

instance Event Interference where
  time = intTime
  turn = intTurn
  getCal = ghInts
  setCal x gh = gh{ ghInts = x }

instance Event AttackResult where
  time = atkTime . arAttack
  turn = atkTurn . arAttack
  getCal = ghAtkResults
  setCal x gh = gh{ ghAtkResults = x }

instance Event Damage where
  time = dmgTime
  turn = dmgTurn
  getCal = ghDmgs
  setCal x gh = gh{ ghDmgs = x }

instance Event Continue where
  time = contTime
  turn = contTurn
  getCal = ghConts
  setCal x gh = gh{ ghConts = x }

instance Event Drop where
  time = dropTime
  turn = dropTurn
  getCal = ghDrops
  setCal x gh = gh{ ghDrops = x }

type Cal a = M.Map Time a
data GameHistory = GH
  { ghTurns :: !(Cal Turn)
  , ghInvs :: !(Cal Invocation)
  , ghAttacks :: !(Cal Attack)
  , ghSnapshots :: !(Cal GameState)
  , ghAtkResults :: !(Cal AttackResult)
  , ghInts :: !(Cal Interference)
  , ghDmgs :: !(Cal Damage)
  , ghConts :: !(Cal Continue)
  , ghDrops :: !(Cal Drop)
  }

data GameFinishedException = GFE GameResult GameInfo
  deriving (Typeable)

newtype GameResult = Winner CharDescr
  deriving (Show, Eq)

data SlotDescr = SD
  { sdChar :: !CharDescr
  , sdSlot :: !Int
  , sdSkill :: !Skill
  }
  deriving (Show, Eq)

data SkillCondition = UseByDefault | DontUseByDefault | NeverUse

data Skill
  = Sk
  { skName :: !String
  , skCosts :: ![PanelKind]
  , skEffect :: !(Maybe (Invocation -> Game ()))
  , skCondition :: !(Game SkillCondition)
  , skReaction :: !(Maybe AttackReaction)
  , skReactionCondition :: !(Attack -> Game SkillCondition)
  , skReactionAgainstSkill :: !(Maybe SkillReaction)
  , skReactionConditionAgainstSkill :: !(Invocation -> Game SkillCondition)
  , skAttrs :: !(S.Set SkillAttribute)
  , skTargets :: !(Maybe (CharDescr -> Game (S.Set SkillTarget)))
  }

type AttackReaction = Invocation -> Int -> Attack -> Game (Int, Bool, Interference -> Game ())
type SkillReaction = Invocation -> Invocation -> Game Bool

instance Show Skill where
  showsPrec _ o = (skName o ++)

instance Eq Skill where
  a == b = skName a == skName b

instance Ord Skill where
  compare a b = compare (skName a) (skName b)

data SkillAttribute
  = AttackingS
  | CounterS
  | MutativeS
  | ObstructiveS
  | ReactiveS
  | DefensiveS
  | RecoveringS
  | ContinuousS
  | StrengtheningS
  deriving (Eq, Ord, Show)

getPlanState :: CharDescr -> Int -> Game (Maybe PlanState)
getPlanState cd n = do
  pps <- getPresentPlans cd
  return $ find ((==n) . psIndex) pps

getPresentPlans :: CharDescr -> Game [PlanState]
getPresentPlans cd = liftM csPlans $ getCS cd

addPlanBind :: CharDescr -> Int -> Int -> Game ()
addPlanBind cd which n = modifyCS cd $ \cs -> cs{ csPlans = f $ csPlans cs }
  where
    f = findAndUpdate "addPlanBind: plan not found" ((==which) . psIndex) add
    add ps@PS{ psBound = b } = ps{ psBound = b + n }

findAndUpdate :: String -> (a -> Bool) -> (a -> a) -> [a] -> [a]
findAndUpdate msg fin upd xs = case break fin xs of
  (_, []) -> error msg 
  (pre, x:post) -> pre ++ [upd x] ++ post

deletePlan :: CharDescr -> Int -> Game ()
deletePlan cd pid = do
  pss <- getPresentPlans cd
  case break ((==pid) . psIndex) pss of
    (_, []) -> fail $ "deletePlan: plan not found"
    (as, _:bs) -> setPSs cd $ as ++ bs

undeletePlan :: CharDescr -> Int -> Game ()
undeletePlan cd pid = do
  pss <- getPresentPlans cd
  plans <- getPlans cd
  hints <- getHints cd
  let ps = initialPS (plans !! pid) (hints !! pid) pid
  case break ((>pid) . psIndex) pss of
    (as, bs) -> setPSs cd $ as ++ [ps] ++ bs

setPSs :: CharDescr -> [PlanState] -> Game ()
setPSs cd pss = modifyCS cd $ \cs -> cs{ csPlans = pss }

getSSs :: CharDescr -> Game [SkillState]
getSSs cd = liftM csSkills $ getCS cd

getSS :: CharDescr -> Int -> Game (Maybe SkillState)
getSS cd n = liftM (find ((==n) . ssIndex)) $ getSSs cd

modifySSs :: CharDescr -> ([SkillState] -> [SkillState]) -> Game ()
modifySSs cd f = modifyCS cd $ \cs -> cs{ csSkills = f $ csSkills cs }

modifySS :: CharDescr -> Int -> (SkillState -> SkillState) -> Game ()
modifySS cd n f = modifySSs cd $ findAndUpdate "modifySS: ss not found" ((==n) . ssIndex) f

addSkillBind :: CharDescr -> Int -> Int -> Game ()
addSkillBind cd n amount = modifySS cd n $ \ss -> ss{ ssBound = amount + ssBound ss }

addSkillWeaken :: CharDescr -> Int -> Int -> Game ()
addSkillWeaken cd n amount = modifySS cd n $ \ss -> ss{ ssWeakened = amount + ssWeakened ss }

deleteSlot :: CharDescr -> Int -> Game ()
deleteSlot cd n = modifySSs cd $ filter ((/=n) . ssIndex)

undeleteSlot :: CharDescr -> Int -> Game ()
undeleteSlot cd n = do
  ch <- char cd
  let ss = initialSS (chSkills ch !! n) n
  modifySSs cd $ insertOn ssIndex ss

setSSs :: CharDescr -> [SkillState] -> Game ()
setSSs cd sss = modifyCS cd $ \cs -> cs{ csSkills = sss }

getSPSs :: CharDescr -> Game [SkillPatState]
getSPSs cd = liftM csSkillPats $ getCS cd

getHSs :: CharDescr -> Game [HeldSkill]
getHSs cd = liftM csSkillsHeld $ getCS cd

setHSs :: CharDescr -> [HeldSkill] -> Game ()
setHSs cd hss = modifyCS cd $ \cs -> cs{ csSkillsHeld = hss }

getCurrentSP :: CharDescr -> Game (Maybe Int)
getCurrentSP cd = liftM csCurrentSP $ getCS cd

withCurrentSP :: CharDescr -> a -> (Int -> Game a) -> Game a
withCurrentSP cd def f = do
  mcsp <- getCurrentSP cd
  case mcsp of
    Nothing -> return def
    Just csp -> f csp

setCurrentSP :: CharDescr -> Int -> Game ()
setCurrentSP cd ind = modifyCS cd $ \cs -> cs{ csCurrentSP = Just ind }

deleteCurrentSP :: CharDescr -> Game ()
deleteCurrentSP cd = modifyCS cd $ \cs -> cs{ csCurrentSP = Nothing }

deleteSP :: CharDescr -> Int -> Game ()
deleteSP cd ind = do
  modifySPSs cd $ filter ((/=ind) . spsIndex)
  mcsp <- getCurrentSP cd
  when (mcsp == Just ind) $ deleteCurrentSP cd

undeleteSP :: CharDescr -> Int -> Game ()
undeleteSP cd ind = do
  ch <- char cd
  let sps = initialSPS (chSkillPats ch !! ind) (chSkillPatNames ch !! ind) ind
  modifySPSs cd $ insertOn spsIndex sps

getSPS :: CharDescr -> Int -> Game (Maybe SkillPatState)
getSPS cd n = do
  cs <- getCS cd
  return $ find ((==n) . spsIndex) $ csSkillPats cs

getSPSByName :: CharDescr -> String -> Game (Maybe SkillPatState)
getSPSByName cd name = liftM (find ((==name) . spsName)) $ getSPSs cd

getPanels :: CharDescr -> Game [Panel]
getPanels cd = withCurrentSP cd [] $ \csp -> do
  sps <- getSPS cd csp
  case sps of
    Nothing -> fail "getPanels: no current SP"
    Just sp -> return $ spsPanels sp

getPanel :: CharDescr -> Int -> Int -> Game (Maybe Panel)
getPanel cd m n = do
  msps <- getSPS cd m
  return $ do
    sps <- msps
    find ((==n) . paIndex) $ spsPanels sps

getPanel' :: CharDescr -> Int -> Int -> Game Panel
getPanel' cd m n = liftM fromJust $ getPanel cd m n

deletePanel :: CharDescr -> Int -> Int -> Game ()
deletePanel cd m n = modifyPanels cd m $ filter ((/=n) . paIndex)

undeletePanel :: CharDescr -> Int -> Int -> Game ()
undeletePanel cd m n = do
  ch <- char cd
  let pa = initialPa (chSkillPats ch !! m !! n) n
  modifyPanels cd m $ insertOn paIndex pa

modifySPSs :: CharDescr -> ([SkillPatState] -> [SkillPatState]) -> Game ()
modifySPSs cd f = modifyCS cd $ \cs -> cs{ csSkillPats = f $ csSkillPats cs }

modifySPS :: CharDescr -> Int -> (SkillPatState -> SkillPatState) -> Game ()
modifySPS cd n f = modifySPSs cd $ findAndUpdate "modifySPS: SPS not found" ((==n) . spsIndex) f

modifyPanels :: CharDescr -> Int -> ([Panel] -> [Panel]) -> Game ()
modifyPanels cd m f = modifySPS cd m $ \sps -> sps{ spsPanels = f $ spsPanels sps }

modifyPanel' :: CharDescr -> Int -> Int -> (Panel -> Panel) -> Game ()
modifyPanel' cd m n f = modifyPanels cd m $ findAndUpdate "modifyPanel: panel not found" ((==n) . paIndex) f

modifyPanel :: CharDescr -> Int -> Int -> [PanelOperation] -> Game ()
modifyPanel _ _ _ [] = return () -- shortcut
modifyPanel cd m n pos = modifyPanel' cd m n $ \x -> foldr applyPO x pos

modifyCurrentPanels :: CharDescr -> [(Int, PanelOperation)] -> Game ()
modifyCurrentPanels cd ops = do
  mcsp <- getCurrentSP cd
  case mcsp of
    Nothing
      | null ops -> return ()
      | otherwise -> fail $ "modifyCurrentPanels: no csp"
    Just csp -> modifyPanels cd csp $ \ps -> snd $ mapAccumL trans (sortOn fst ops) ps
  where
    trans ops pa = (rest, foldr applyPO pa cur)
      where
        (cur, rest) = takeSpan (paIndex pa) ops

takeSpan :: (Ord a) => a -> [(a, b)] -> ([b], [(a, b)])
takeSpan _ [] = ([], [])
takeSpan k all@((a, b):rest)
  | a < k = (c, r)
  | a == k = (b : c, r)
  | a > k = ([], all)
  where
    (c, r) = takeSpan k rest

applyPO :: PanelOperation -> Panel -> Panel
applyPO po pa = case po of
  SetPoison n -> pa{ paPoisoned = n }
  AddPoison n -> pa{ paPoisoned = n + paPoisoned pa }
  SetConfuse n -> pa{ paConfused = n }
  AddConfuse n -> pa{ paConfused = n + paConfused pa }
  SetBind n -> pa{ paBound = n }
  AddBind n -> pa{ paBound = n + paBound pa }
  SetKind k -> pa{ paKind = k }
  SetUsed u -> pa{ paUsed = u }
  SetEnsnared e -> pa{ paEnsnared = e }
  SetBaseKind k -> pa{ paBaseKind = k }

data PanelOperation
  = SetPoison Int
  | AddPoison Int
  | SetConfuse Int
  | AddConfuse Int
  | SetBind Int
  | AddBind Int
  | SetKind PanelKind
  | SetUsed Bool
  | SetEnsnared Bool
  | SetBaseKind PanelKind

modifyHS :: CharDescr -> Int -> (HeldSkill -> HeldSkill) -> Game ()
modifyHS cd n f = getHSs cd >>= setHSs cd . update n f

getSarasouju :: CharDescr -> Game Int
getSarasouju cd = liftM csSarasouju $ getCS cd

modifySarasouju :: CharDescr -> (Int -> Int) -> Game ()
modifySarasouju cd f = modifyCS cd $ \cs -> cs{ csSarasouju = f $ csSarasouju cs }

getAngelFire :: CharDescr -> Game Int
getAngelFire cd = liftM csAngelFire $ getCS cd

modifyAngelFire :: CharDescr -> (Int -> Int) -> Game ()
modifyAngelFire cd f = modifyCS cd $ \cs -> cs{ csAngelFire = f $ csAngelFire cs }

update :: Int -> (a -> a) -> [a] -> [a]
update n f xs = case splitAt n xs of
  (a, z:b) -> a ++ [f z] ++ b

replace :: Int -> a -> [a] -> [a]
replace n x xs = update n (const x) xs

getHints :: CharDescr -> Game [PlanHint]
getHints cd = liftM chPlanHints $ char cd

getPlans :: CharDescr -> Game [Plan]
getPlans cd = liftM chPlans $ char cd

modifyCS :: CharDescr -> (CharState -> CharState) -> Game ()
modifyCS First f = modifyGS $ \gs -> gs{ gsChar0 = f (gsChar0 gs) }
modifyCS Second f = modifyGS $ \gs -> gs{ gsChar1 = f (gsChar1 gs) }

getGS :: Game GameState
getGS = do
  rg <- asks giState
  liftIO $ readIORef rg

modifyGS :: (GameState -> GameState) -> Game ()
modifyGS modf = do
  rgs <- asks giState
  liftIO $ modifyIORef rgs modf

getGH :: Game GameHistory
getGH = do
  rgh <- asks giHistory
  liftIO $ readIORef rgh

modifyGH :: (GameHistory -> GameHistory) -> Game ()
modifyGH modf = do
  rgh <- asks giHistory
  liftIO $ modifyIORef rgh modf

getCalendar :: (Event a) => Game (Cal a)
getCalendar = liftM getCal getGH

getTurns :: Game (Cal Turn)
getTurns = liftM ghTurns getGH

getSnapshots :: Game (Cal GameState)
getSnapshots = liftM ghSnapshots getGH

getNextTime :: Game Time
getNextTime = liftM gsTimeSource getGS

getCS :: CharDescr -> Game CharState
getCS First = do
  GS{ gsChar0 = cs } <- getGS
  return cs
getCS Second = do
  GS{ gsChar1 = cs } <- getGS
  return cs

char :: CharDescr -> Game Character
char First = asks giChar0
char Second = asks giChar1

getParadiseLost :: Game Int
getParadiseLost = liftM gsParadiseLost getGS

addParadiseLost :: Int -> Game ()
addParadiseLost n = modifyGS $ \gs -> gs{ gsParadiseLost = n + gsParadiseLost gs }

registerReaction :: Invocation -> AttackReaction -> Invocation -> Game ()
registerReaction inv re self = modifyGS $ \gs -> gs{ gsReactions = f $ gsReactions gs }
  where
    f = M.insertWith (++) inv [RI re self]

getReactions :: Invocation -> Game [(AttackReaction, Invocation)]
getReactions inv = liftM (reverse . map unRI . fromMaybe [] . M.lookup inv . gsReactions) getGS
  where
    unRI (RI x y) = (x, y)

vitToMaxHp :: Int -> Int
vitToMaxHp vit = sum [1 .. vit]

getHp :: CharDescr -> Game Int
getHp cd = liftM csHp $ getCS cd

setHp :: CharDescr -> Int -> Game ()
setHp cd hp = modifyCS cd $ \cs -> cs{ csHp = hp }

getVit :: CharDescr -> Game Int
getVit cd = liftM csVit $ getCS cd

setVit :: CharDescr -> Int -> Game ()
setVit cd n = modifyCS cd $ \cs -> cs{ csVit = n }

getBaseVit :: CharDescr -> Game Int
getBaseVit cd = liftM csBaseVit $ getCS cd

setBaseVit :: CharDescr -> Int -> Game ()
setBaseVit cd n = modifyCS cd $ \cs -> cs{ csBaseVit = n }

getRefl :: CharDescr -> Game Int
getRefl cd = liftM csRefl $ getCS cd

setRefl :: CharDescr -> Int -> Game ()
setRefl cd n = modifyCS cd $ \cs -> cs{ csRefl = n }

getInt :: CharDescr -> Game Int
getInt cd = liftM csInt $ getCS cd

setInt :: CharDescr -> Int -> Game ()
setInt cd n = modifyCS cd $ \cs -> cs{ csInt = n }

getCurrentTurn :: Game Turn
getCurrentTurn = liftM (fromMaybe (error "getCurrentTurn: no current turn") . gsCurrentTurn) getGS

setCurrentTurn :: Turn -> Game ()
setCurrentTurn t = modifyGS $ \gs -> gs{ gsCurrentTurn = Just t }

isSuddenDeath :: Game Bool
isSuddenDeath = asks giIsSuddenDeath

gameFinally :: Game a -> Game () -> Game a
gameFinally x y = do
  gi <- ask
  liftIO $ runGame x gi `finally` runGame y gi

withGS :: GameState -> Game a -> Game a
withGS newgs x = do
  rgs <- asks giState
  oldgs <- liftIO $ readIORef rgs
  liftIO $ writeIORef rgs newgs
  x `gameFinally` (liftIO $ writeIORef rgs oldgs)

newUnique :: Game Unique
newUnique = do
  ru <- asks giUniqueSource
  old <- liftIO $ readIORef ru
  liftIO $ writeIORef ru (succ old)
  return old

mes :: String -> Game ()
mes str = do
  out <- asks giOutput
  case out of
    Just o -> liftIO $ o str
    Nothing -> return ()

tickTime :: Game Time
tickTime = do
  join $ asks giTickHook
  rgs <- asks giState
  gs <- liftIO $ readIORef rgs
  let old = gsTimeSource gs
  liftIO $ writeIORef rgs $ gs{ gsTimeSource = succ old }
  modifyGH $ \gh -> gh{ ghSnapshots = M.insert old gs $ ghSnapshots gh }
  return old

recordEvent :: (Event a) => a -> Game ()
recordEvent evt = modifyGH $ \gh -> setCal (M.insert (time evt) evt $ getCal gh) gh

recordTurn :: Turn -> Game ()
recordTurn t = modifyGH $ \gh -> gh{ ghTurns = M.insert (turnTime t) t $ ghTurns gh }

initialState :: Character -> Character -> GameState
initialState x y = GS
  { gsCurrentTurn = Nothing
  , gsChar0 = initialCS x
  , gsChar1 = initialCS y
  , gsTimeSource = initialTime
  , gsParadiseLost = 0
  , gsReactions = M.empty
  }
  where
    initialCS ch = CS
      { csHp = vitToMaxHp $ chVit ch
      , csVit = chVit ch
      , csBaseVit = chVit ch
      , csRefl = chReflex ch
      , csInt = chInt ch
      , csSkillPats = zipWith3 initialSPS (chSkillPats ch) (chSkillPatNames ch) [0..]
      , csSkills = zipWith initialSS (chSkills ch) [0..]
      , csCurrentSP = if null (chSkillPats ch) then Nothing else Just 0
      , csPlans = zipWith3 initialPS (chPlans ch) (chPlanHints ch) [0..]
      , csSkillsHeld = []
      , csSarasouju = 0
      , csAngelFire = 0
      }

initialSS :: Skill -> Int -> SkillState
initialSS sk idx = SS
  { ssBound = 0
  , ssWeakened = 0
  , ssIndex = idx
  , ssSkill = sk
  }

initialPS :: Plan -> PlanHint -> Int -> PlanState
initialPS plan hint idx = PS
  { psBound = 0
  , psPlan = plan
  , psHint = hint
  , psIndex = idx
  }

initialSPS :: SkillPattern -> String -> Int -> SkillPatState
initialSPS pats name idx = SPS
  { spsPanels = zipWith initialPa pats [0..]
  , spsName = name
  , spsIndex = idx
  }

initialPa :: PanelKind -> Int -> Panel
initialPa pk idx = Pa
  { paPoisoned = 0
  , paConfused = 0
  , paBound = 0
  , paKind = pk
  , paUsed = False
  , paEnsnared = False
  , paBaseKind = pk
  , paOriginalKind = pk
  , paIndex = idx
  }

initialHistory :: GameHistory
initialHistory = GH
  { ghTurns = M.empty
  , ghInvs = M.empty
  , ghAttacks = M.empty
  , ghSnapshots = M.empty
  , ghAtkResults = M.empty
  , ghInts = M.empty
  , ghDmgs = M.empty
  , ghConts = M.empty
  , ghDrops = M.empty
  }

newGameInfo :: Bool -> RndGen -> (Maybe (String -> IO ())) -> Game () -> Character -> Character -> IO GameInfo
newGameInfo sd gen out hook x y = do
  rs <- newIORef $ initialState x y
  rh <- newIORef initialHistory
  ru <- newIORef initialUnique
  rg <- newIORef gen
  return GI
    { giChar0 = x
    , giChar1 = y
    , giState = rs
    , giIsSuddenDeath = sd
    , giUniqueSource = ru
    , giOutput = out
    , giHistory = rh
    , giTickHook = hook
    , giRndGen = rg
    }

getRndGen :: Game RndGen
getRndGen = do
  r <- asks giRndGen
  liftIO $ readIORef r

newRndGen :: Game RndGen
newRndGen = do
  r <- asks giRndGen
  gen <- liftIO $ readIORef r
  let (g0, g1) = split gen
  liftIO $ writeIORef r g0
  return g1

rnd :: Int -> Game Int
rnd n = do
  rg <- asks giRndGen
  g <- liftIO $ readIORef rg
  case next g of
    (v, g') -> do
      liftIO $ writeIORef rg g'
      return $ v `mod` n

isOutputPresent :: Game Bool
isOutputPresent = asks $ isJust . giOutput

getGSFromGI :: GameInfo -> IO GameState
getGSFromGI gi = readIORef $ giState gi

sortOn :: (Ord a) => (b -> a) -> [b] -> [b]
sortOn f xs = sortBy cmp xs
  where
    cmp x y = compare (f x) (f y)

insertOn :: (Ord a) => (b -> a) -> b -> [b] -> [b]
insertOn f x xs = insertBy cmp x xs
  where
    cmp x y = compare (f x) (f y)
