module GameUtils where

import Game

import Control.Monad.Reader
import Data.Maybe
import Data.List
import qualified Data.Set as S
import Data.Dynamic
import Control.Exception

chIsValid :: Character -> Bool
chIsValid Ch{ chReflex = reflex, chInt = int, chSkillPats = pats, chSkillPatNames = spns, chSkills = skills, chPlans = plans, chPlanHints = phs }
  = all (<=reflex) (map length pats) &&
      skillFair &&
      length plans <= int &&
      length plans == length phs &&
      length pats <= int &&
      length spns == length pats &&
      nodup spns
  where
    skillFair = length level2Skills + max 0 (length level1Skills - nFreeSkill) <= int
    level2Skills = filter ((==2) . skLevel) skills
    level1Skills = filter ((==1) . skLevel) skills
    nodup xs = length xs == length (nub xs)

skLevel :: Skill -> Int
skLevel = length . skCosts

skIsSword :: Skill -> Bool
skIsSword = elem Sword . skCosts

skIsFire :: Skill -> Bool
skIsFire = elem Fire . skCosts

skIsMagic :: Skill -> Bool
skIsMagic = not . skIsSword

skIsAttacking, skIsCounter, skIsMutative, skIsObstructive,
  skIsReactive, skIsDefensive, skIsRecovering, skIsContinuous :: Skill -> Bool
skIsAttacking = skHasProp AttackingS
skIsCounter = skHasProp CounterS
skIsMutative = skHasProp MutativeS
skIsObstructive = skHasProp ObstructiveS
skIsReactive = skHasProp ReactiveS
skIsDefensive = skHasProp DefensiveS
skIsRecovering = skHasProp RecoveringS
skIsContinuous = skHasProp ContinuousS
skIsStrengthening = skHasProp StrengtheningS

skHasProp :: SkillAttribute -> Skill -> Bool
skHasProp sa sk = S.member sa $ skAttrs sk

paUsable :: Panel -> Bool
paUsable Pa{ paBound = b, paUsed = u } = not u && b == 0

skillReady :: CharDescr -> Skill -> Game Bool
skillReady cd skill = do
  c0 <- hasSkill cd skill
  c1 <- canAfford cd skill
  sss <- getSSs cd
  let c2 = any usable sss
  return $ c0 && c1 && c2
  where
    usable ss = ssSkill ss == skill && ssBound ss == 0

-- the result may contain duplicate skills
skillsReady :: CharDescr -> Game [Skill]
skillsReady cd = liftM (map (ssSkill . fst) . filter snd) $ skillsReadiness cd

skillsReadiness :: CharDescr -> Game [(SkillState, Bool)]
skillsReadiness cd = do
  sss <- getSSs cd
  panels <- getPanels cd
  let avail_c = countKinds $ map paKind $ filter paUsable $ panels
  seq avail_c $ do
    let
      ok ss
        | ssBound ss == 0 && costLessEq (skCosts $ ssSkill ss) avail_c
          = (ss, True)
        | otherwise = (ss, False)
    return $ map ok sss

costLessEq :: [PanelKind] -> KindCount -> Bool
costLessEq [] _ = True
costLessEq [k] c = kcRef k c >= 1
costLessEq [k0, k1] c
  | k0 == k1 = kcRef k0 c >= 2
  | otherwise = kcRef k0 c >= 1 && kcRef k1 c >= 1
costLessEq ks c = countLe (countKinds ks) c

kcRef :: PanelKind -> KindCount -> Int
kcRef Fire (KindCount f _ _ _ _) = f
kcRef Water (KindCount _ wa _ _ _) = wa
kcRef Wind (KindCount _ _ wi _ _) = wi
kcRef Earth (KindCount _ _ _ e _) = e
kcRef Sword (KindCount _ _ _ _ s) = s

hasSkill :: CharDescr -> Skill -> Game Bool
hasSkill cd skill = do
  sss <- getSSs cd
  return $ any ((==skill) . ssSkill) sss

canAfford :: CharDescr -> Skill -> Game Bool
canAfford cd sk = do
  panels <- getPanels cd
  let costsC = countKinds $ skCosts sk
  let availC = countKinds $ map paKind $ filter paUsable panels
  return $ countLe costsC availC

data KindCount = KindCount {-# UNPACK #-} !Int {-# UNPACK #-} !Int {-# UNPACK #-} !Int {-# UNPACK #-} !Int {-# UNPACK #-} !Int

countKinds :: [PanelKind] -> KindCount
countKinds pks = loop 0 0 0 0 0 pks
  where
    loop f wa wi e s _
      | f `seq` wa `seq` wi `seq` e `seq` s `seq` False = undefined
    loop f wa wi e s [] = KindCount f wa wi e s
    loop f wa wi e s (k:r) = case k of
      Fire -> loop (f+1) wa wi e s r
      Water -> loop f (wa+1) wi e s r
      Wind -> loop f wa (wi+1) e s r
      Earth -> loop f wa wi (e+1) s r
      Sword -> loop f wa wi e (s+1) r

countLe :: KindCount -> KindCount -> Bool
countLe (KindCount a b c d e) (KindCount a1 b1 c1 d1 e1)
  = a <= a1 && b <= b1 && c <= c1 && d <= d1 && e <= e1

mesC :: CharDescr -> String -> Game ()
mesC cd str = do
  ch <- getPrettyName cd
  mes $ ch ++ str

mesPa :: CharDescr -> Int -> Int -> String -> Game ()
mesPa cd m n str = mesC cd $ "のスキルパターン" ++ show m ++ "のパネル" ++ show n ++ str

getPrettyName :: CharDescr -> Game String
getPrettyName cd = do
  name <- cname cd
  nameo <- cname $ opponent cd
  return $ if name == nameo then name ++ "(" ++ show (charDescrToIndex cd) ++ ")" else name
  where
    cname cd1 = liftM (fromMaybe (show cd1)) $ getName cd1

kindPanels :: CharDescr -> PanelKind -> Game [(Int, Int)]
kindPanels cd kind = do
  spss <- getSPSs cd
  return $ do
    sps <- spss
    pa <- spsPanels sps
    guard $ paKind pa == kind
    return (spsIndex sps, paIndex pa)

choose :: [a] -> Game a
choose [] = fail "choose []"
choose xs = do
  v <- rnd $ length xs
  return $ xs !! v

decrVit :: CharDescr -> Int -> Game ()
decrVit cd n
  | n > 0 = do
    mesC cd $ "の体力が" ++ show n ++ "点減少した"
    cvit <- getVit cd
    cbase <- getBaseVit cd
    let newvit = cvit - n
    let newbase = min cbase newvit
    setVit cd newvit
    setBaseVit cd newbase
    sara <- getSarasouju cd
    when (newvit <= 0 && sara > 0) $ lose cd
  | otherwise = return ()

finish :: GameResult -> Game ()
finish r = do
  gi <- ask
  liftIO $ throwIO $ DynException $ toDyn $ GFE r gi

win :: CharDescr -> Game ()
win w = do
  mesC w $ "が勝利した"
  finish $ Winner w

lose :: CharDescr -> Game ()
lose = win . opponent

hsCosts :: HeldSkill -> [(Int, Int)]
hsCosts = invCosts . hsSource

skillSlots :: CharDescr -> Skill -> Game [Int]
skillSlots cd skill = liftM (map ssIndex . filter ((==skill) . ssSkill)) $ getSSs cd

charDescrToIndex :: CharDescr -> Int
charDescrToIndex First = 0
charDescrToIndex Second = 1

opponent :: CharDescr -> CharDescr
opponent First = Second
opponent Second = First

getFirstToMove :: Game CharDescr
getFirstToMove = liftM turnFirstToMove getCurrentTurn

getTurnCount :: Game Int
getTurnCount = liftM turnNumber getCurrentTurn

getName :: CharDescr -> Game (Maybe String)
getName cd = liftM chName $ char cd

getCurrentSP' :: CharDescr -> Game Int
getCurrentSP' cd = do
  mcsp <- getCurrentSP cd
  maybe (fail "getCurrentSP': no csp") return mcsp

type Delay = (Time, Game ())

performDelays :: [Delay] -> Game ()
performDelays = sequence_ . map snd . sortOn fst

chooseMaybe :: [a] -> Game (Maybe a)
chooseMaybe [] = return Nothing
chooseMaybe xs = liftM Just $ choose xs

concatMapM :: (Monad m) => (a -> m [b]) -> [a] -> m [b]
concatMapM f = liftM concat . mapM f

invIsReaction :: Invocation -> Bool
invIsReaction = invTypeIsReaction . invType

invTypeIsReaction :: InvocationType -> Bool
invTypeIsReaction ReactionI{} = True
invTypeIsReaction ReactionAgainstSkillI{} = True
invTypeIsReaction _ = False

evalSkillCondition :: Bool -> SkillCondition -> Bool
evalSkillCondition force sc = case sc of
  UseByDefault -> True
  DontUseByDefault -> force
  NeverUse -> False
