{-# LANGUAGE Rank2Types #-}
module GlobalEnv(globalEnv) where

import Game
import GameUtils
import ExprUtils
import {-# SOURCE #-} AllSkills
import Plan

import Data.Char
import Control.Monad
import qualified Data.Set as S
import qualified Data.Map as M
import Data.List
import qualified Data.Foldable as Fold
import Data.Monoid(mempty, mappend, mconcat)
import Data.Maybe
import Control.Monad.Reader
import System.Random

globalEnv :: Env
globalEnv = M.fromList $
    map val
    -- basic constants
    [ ("self", RChV Self)
    , ("opp", RChV Opp)
    , ("true", BoolV True)
    , ("false", BoolV False)
    , ("nil", NilV)
    -- just for testing
    --, ("avenger", SkillV avenger)
    -- other non-functions
    , ("all_kinds", SetV $ S.fromList $ map PanelKindV [minBound .. maxBound])
    , ("all_skills", SetV $ S.fromList $ map SkillV allSkills)
    , ("nop", ActionV mempty)
    , ("opening", TimeV initialTime)
    ] ++
    map thunk
    -- global state query
    [ ("first_to_move", liftM RChV getCurrentFirstInRel)
    , ("first", liftM (BoolV . (==Self)) getCurrentFirstInRel)
    , ("second", liftM (BoolV . (/=Self)) getCurrentFirstInRel)
    , ("turn_count", liftM (intV . (1+)) $ liftGame getTurnCount)
    , ("current_turn", liftM TurnV $ liftGame getCurrentTurn)
    , ("current_invocation", liftM (maybeNil InvocationV . ecCurrentInv) getContext)
    , ("current_attack", liftM (maybeNil AttackV . ecCurrentAtk) getContext)
    -- history variables
    , ("all_invocations", allEvents InvocationV)
    , ("all_attacks", allEvents AttackV)
    , ("all_blocks", allEvents InterferenceV)
    , ("all_turns", liftM (makeEventSetR TurnV) $ liftGame getTurns)
    , ("all_damages", allEvents DamageV)
    , ("all_continuations", allEvents ContinuationV)
    , ("all_drops", allEvents DropV)
    -- rule queries
    , ("sudden_death", liftGame $ liftM BoolV isSuddenDeath)
    -- global random number generator
    , ("init_g", liftM (rndgenV . ecRndGen) getContext)
    ] ++
    skillVars ++
    panelKinds ++
    map pfn
    [
      -- basic fuctions
      pId
    , pConst
    , pNot
    , pEq
    , pEven
    , pOdd
    , pPow
      -- character information
    , pCharName
    , pVitality
    , pIntelligence
    , pReflexes
    , pHP
    , pMaxHP
    , pHasSkill
    , pSkills
    , pReady
    , pSkillsReady
    , pSkillSlots
    , pPatterns
    , pCurrentPatternNumber
    , pCurrentPattern
    , pCurrentPanels
    , pCurrentPanelKinds
    , pHeldInvocations
    , pHolding
    , pPlans
    -- skills
    , pSkillName
    , pCost
    , skProp "is_sword" skIsSword
    , skProp "is_fire" skIsFire
    , skProp "is_magic" skIsMagic
    , skProp "is_attack" skIsAttacking
    , skProp "is_counter" skIsCounter
    , skProp "is_mutative" skIsMutative
    , skProp "is_obstructive" skIsObstructive
    , skProp "is_reactive" skIsReactive
    , skProp "is_defensive" skIsDefensive
    , skProp "is_recovery" skIsRecovering
    , skProp "is_continuous" skIsContinuous
    , skProp "is_strengthening" skIsStrengthening
    , pPossibleTargets
    , pLevel
    , pElement
    -- skill slots
    , pSlotSkill
    , pWeaken
    , pIsWeakened
    -- panels
    , pKind
    , pPoison
    , pConfuse
    , pIsPoisoned
    , pIsConfused
    , pIsUsed
    , pIsEnsnared
    -- plans
    , pPlanCode
    -- overloaded "boundness" query
    , pBind
    , pIsBound
    -- overloaded owner function
    , pOwner
    -- overloaded existence query
    , pExists
    -- actions
    , pDontUse
    , pPrioritize
    , pPrioritizeBelow
    , pPrioritizeAbove
    , pPrioritizeBottom
    , pUse
    , pTarget
    , pTargetWithPriorities
    , pSwitchTo
    , pReserve
    , pHold
    , pSetReactionCount
    , pModifyPlan
    , pSeq
    , pSequence
    -- sets
    , pUnion
    , pUnions
    , pIntersection
    , pDifference
    , pInsert
    , pDelete
    , pImage
    -- overloaded functions between sets and lists
    , pMember
    , pSize
    , pEmpty
    , pFilter
    , pAny
    , pAll
    , pCount
    , pMax
    , pMaxOn
    , pMin
    , pMinOn
    , pToList
    , pToSet
    -- invocations
    , pInvocationSlot
    , pInvocationSkill
    , pInvocationCost
    , pInvocationType
    , pIsAction
    , pIsReaction
    , pInvocationCause
    , pInvoker
    , pTurnInvocation
    -- overloaded operations over events
    , pTime
    , pTurn
    -- attacks
    , pAttackInvocation
    , pAttackSkill
    , pAttackCause
    , pClaimedDamage
    , pAttackIsSword
    , pAttackIsFire
    , pAttacker
    , pAttackee
    , pDamageAmount
    , pDamage
    , pIsNegated
    , pBlocks
    , pTurnAttacks
    , pTurnAttacksFrom
    -- blocks
    , pBlockTarget
    , pInputDamage
    , pOutputDamage
    , pBlockNegate
    , pBlockInvocation
    , pBlockSkill
    , pBlockType
    , pTurnBlocks
    -- turns
    , pBeginning
    , pEnd
    , pNext
    , pPrev
    , pTurnNumber
    , pNumberTurn
    -- debugging
    , pDebugOut
    -- lists
    , pFind
    , pAppend
    , pFoldr
    , pHead
    , pTail
    , pCons
    , pSort
    , pSortOn
    , pReverse
    , pMap
    -- damages
    , pDamageValue
    , pSufferer
    , pDamageType
    , pDamageAttack
    , pTurnDamages
    -- skill patterns
    , pPanels
    , pPatternName
    , pPatternNumber
    , pPatternByName
    -- skill continuations
    , pContinuedInvocation
    , pTurnContinuations
    -- skill drops
    , pDroppedInvocation
    , pDropType
    , pTurnDrops
    -- strings
    , pUnpack
    , pPack
    -- RNGs
    , pSplit
    , pRandomInt
    , pChoose
    ]
pSplit = binRandPrim "split" $ \gen x -> do
  n <- toInt "split" x
  if n < 0
    then err $ "split: negative count: " ++ show n
    else return $ ListV $ take n $ map rndgenV $ unfoldr (Just . split) gen
pRandomInt = binRandPrim "random_int" $ \gen x -> do
  n <- toInteg "random_int" x
  if n <= 0
    then err $ "random_int: not positive: " ++ show n
    else return $ IntegerV $ fst $ randomR (0, n-1) gen
pChoose = binRandPrim "choose" $ \gen x -> do
  vs <- toList' "choose" x
  if null vs
    then err $ "choose: empty list"
    else return $ vs !! (fst $ randomR (0, length vs - 1) gen)

pUnpack = unPrim0 "unpack" $ \x -> do
  str <- toString "unpack" x
  return $ ListV $ map (intV . ord) str
pPack = unListPrim "pack" $ liftM StringV . packToString

pDroppedInvocation = unDropPrim "dropped_invocation" $ return . InvocationV . dropInv
pDropType = unDropPrim "drop_type" $ \drp -> return $ StringV $ case dropReason drp of
  DropByWill{} -> "will"
  DropByRoll{} -> "roll"
  DropByNegation{} -> "negation"
  DropByLostPanel{} -> "lost panel"
  DropByLostSkill{} -> "lost skill"
pTurnDrops = turnEvents "turn_drops" DropV (invControler . dropInv)

pContinuedInvocation = unPrim' "continued_invocation" $ \cd x -> do
  cont <- toCont "continued_invocation" x
  return $ InvocationV $ contInv cont
pTurnContinuations = turnEvents "turn_continuations" ContinuationV (invControler . contInv)

pPanels = unPatPrim "panels" $ \cd1 sps -> return $ let
    makePanel = PanelV cd1 (spsIndex sps)
  in ListV $ map (makePanel . paIndex) $ spsPanels sps
pPatternName = unPatPrim "pattern_name" $ \cd1 sps -> return $ StringV $ spsName sps
pPatternNumber = unPatPrim "pattern_number" $ \cd1 sps -> return $ intV $ spsIndex sps
pPatternByName = binCharPrim "pattern_by_name" $ \cd1 x -> do
  name <- toString "pattern_by_name" x
  sps <- getSPSByName cd1 name
  return $ maybeNil (SkillPatternV cd1 . spsIndex) sps

pDamageValue = unDmgPrim "damage_value" $ return . intV . dmgAmount
pSufferer = unPrim' "sufferer" $ \cd x -> do
  dmg <- toDmg "sufferer" x
  return $ RChV $ toRel cd $ dmgSufferer dmg
pDamageType = unDmgPrim "damage_type" $ \dmg -> return $ StringV $ case dmgReason dmg of
  ByAttack{} -> "attack"
  ByPoison{} -> "poison"
pDamageAttack = unDmgPrim "damage_attack" $ \dmg -> return $ case dmgReason dmg of
  ByAttack atk -> AttackV atk
  _ -> NilV
pTurnDamages = turnEvents "turn_damages" DamageV dmgSufferer

pFind = binListPrim "find" $ \x lst -> do
  f <- toFunction "find" x
  r <- findM (applyPred "find" f) lst
  return $ maybeNil id r
pAppend = binListPrim "append" $ \x lst -> do
  xl <- toList "append" x
  return $ ListV $ xl ++ lst
pFoldr = terPrim "foldr" $ \x z y -> do
  f <- toFunction "foldr" x
  lst <- toList "foldr" y
  let f' a b = applyFun2 f a b
  foldrM f' z lst
pHead = unListPrim "head" $ \lst -> return $ case lst of
  [] -> NilV
  x:_ -> x
pTail = unListPrim "tail" $ return . ListV . drop 1
pCons = binListPrim "cons" $ \v lst -> return $ ListV $ v:lst
pSort = unFoldPrim "sort" $ \xs -> liftM ListV $ sortByM (valueCompare "sort") xs
pSortOn = binFoldPrim "sort_on" $ \f xs -> let
  cmp x y = do
    xv <- applyFun1 f x
    yv <- applyFun1 f y
    valueCompare "sort_on" xv yv
  in liftM ListV $ sortByM cmp xs
pReverse = unListPrim "reverse" $ return . ListV . reverse
pMap = binFoldPrim "map" $ \fn xs -> liftM ListV $ mapM (applyFun1 fn) xs

pDebugOut = binPrim0 "debug_out" $ \x y -> do
  msg <- toString "debug_out" x
  mes $ msg ++ ": " ++ show y
  return y

pBeginning = unTurnPrim "beginning" $ return . TimeV . turnTime
pEnd = unTurnPrim "end" $ liftM TimeV . turnEndTimeE
pNext = unTurnPrim "next" $ liftM (maybeNil TurnV) . liftGame . numberTurn . succ . turnNumber
pPrev = unTurnPrim "prev" $ liftM (maybeNil TurnV) . liftGame . numberTurn . pred . turnNumber
pTurnNumber = unTurnPrim "turn_number" $ return . intV . (1+) . turnNumber
pNumberTurn = unIntegPrim "number_turn" $ \x -> case integerToInt (x-1) of
  Just v -> liftM (maybeNil TurnV) $ numberTurn v
  _ -> return NilV

pBlockTarget = unIntfPrim "block_target" $ return . AttackV . intTarget
pInputDamage = unIntfPrim "input_damage" $ return . intV . intInput
pOutputDamage = unIntfPrim "output_damage" $ return . intV . intOutput
pBlockNegate = unIntfPrim "block_negate" $ return . BoolV . intNegated
pBlockInvocation = unIntfPrim "block_invocation" $ return . InvocationV . intInv
pBlockSkill = unIntfPrim "block_skill" $ return . SkillV . invSkill . intInv
pBlockType = unIntfPrim "block_type" $ \int -> return $ StringV $ case intType int of
  InvokedI -> "invoked"
  HeldI -> "held"
pTurnBlocks = turnEvents "turn_blocks" InterferenceV (invControler . intInv)

pAttackInvocation = unAttackPrim "attack_invocation" $ return . InvocationV . atkInv
pAttackSkill = unAttackPrim "attack_skill" $ return . SkillV . invSkill . atkInv
pAttackCause = unAttackPrim "attack_cause" $ return . maybeNil InterferenceV . atkCause
pClaimedDamage = unAttackPrim "claimed_damage" $ return . intV . atkClaimedDmg
pAttackIsSword = unAttackPrim "attack_is_sword" $ return . BoolV . skIsSword . invSkill . atkInv
pAttackIsFire = unAttackPrim "attack_is_fire" $ return . BoolV . skIsFire . invSkill . atkInv
pAttacker = unPrim' "attacker" $ \cd x -> do
  atk <- toAttack "attacker" x
  return $ RChV $ toRel cd $ invControler $ atkInv atk
pAttackee = unPrim' "attackee" $ \cd x -> do
  atk <- toAttack "attackee" x
  return $ RChV $ toRel cd $ atkTarget atk
pDamageAmount = unAttackPrim "damage_amount" $ liftM (maybeNil (intV . fromMaybe 0 . liftM dmgAmount . arDamage)) . atkResult
pDamage = unAttackPrim "damage" $ \atk -> do
  mr <- atkResult atk
  return $ maybeNil DamageV $ do
    r <- mr
    arDamage r
pIsNegated = unAttackPrim "is_negated" $ liftM (maybeNil (BoolV . arNegated)) . atkResult
pBlocks = unAttackPrim "blocks" $ \atk -> do
  ints <- turnSlice (atkTurn atk) =<< getCalendar
  return $ makeEventSet (M.filter ((==atk) . intTarget) ints) InterferenceV
pTurnAttacks = turnEvents "turn_attacks" AttackV (invControler . atkInv)
pTurnAttacksFrom = binPrim' "turn_attacks_from" $ \cd x y -> do
  rch <- toRch "turn_attacks_from" x
  t <- toTurn "turn_attacks_from" y
  let cd1 = rel cd rch
  evts <- getCalendar
  let from_cd1 atk = atkTarget atk == cd && invControler (atkInv atk) == cd1
  sl <- liftM (M.filter from_cd1) $ turnSlice t evts
  return $ makeEventSet sl AttackV

pTurnInvocation = turnEvents "turn_invocations" InvocationV invControler
pInvoker = unInvPrim "invoker" $ \cd inv -> return $ RChV $ toRel cd (invControler inv)
pInvocationSlot = unInvPrim "invocation_slot" $ \cd inv -> return $ SkillSlotV (invControler inv) (invSlot inv)
pInvocationSkill = unInvPrim "invocation_skill" $ \cd inv -> return $ SkillV $ invSkill inv
pInvocationCost = unInvPrim "invocation_cost" $ \cd inv ->
  return $ SetV $ S.fromList $ map (\(m, n) -> PanelV (invControler inv) m n) $ invCosts inv
pInvocationType = unInvPrim "invocation_type" $ \cd inv -> return $ StringV $ case invType inv of
  ActionI -> "action"
  ReactionI{} -> "reaction"
  ReactionAgainstSkillI{} -> "skill reaction"
  ReleaseI{} -> "release"
pIsAction = unInvPrim "is_action" $ \cd inv -> return $ BoolV $ invType inv == ActionI
pIsReaction = unInvPrim "is_reaction" $ \cd inv -> return $ BoolV $ invType inv /= ActionI
pInvocationCause = unInvPrim "invocation_cause" $ \cd inv -> case invType inv of
  ActionI -> return NilV
  ReactionI atk -> return $ AttackV atk
  ReactionAgainstSkillI inv1 -> return $ InvocationV inv1
  ReleaseI inv1 -> return $ InvocationV inv1

pTime = unPrim "time" $ \x -> case x of
  InvocationV inv -> f inv
  AttackV atk -> f atk
  InterferenceV int -> f int
  DamageV dmg -> f dmg
  ContinuationV cont -> f cont
  DropV drp -> f drp
  _ -> typeError "time" x "event"
  where
    f x = return $ TimeV $ time x
pTurn = unPrim "turn" $ \x -> case x of
  InvocationV inv -> f inv
  AttackV atk -> f atk
  InterferenceV int -> f int
  DamageV dmg -> f dmg
  ContinuationV cont -> f cont
  DropV drp -> f drp
  _ -> typeError "turn" x "event"
  where
    f x = return $ TurnV $ turn x

pMax = unFoldPrim "max" $ \xs -> liftM (maybeNil id) $ maximumByM (valueCompare "max") xs
pMin = unFoldPrim "min" $ \xs -> liftM (maybeNil id) $ maximumByM (flip (valueCompare "min")) xs
pMaxOn = binFoldPrim "max_on" $ \f xs -> liftM (maybeNil id) $ maximumByM (valueCompareOn "max_on" f) xs
pMinOn = binFoldPrim "min_on" $ \f xs -> liftM (maybeNil id) $ maximumByM (flip $ valueCompareOn "min_on" f) xs
valueCompareOn name f a b = do
  va <- applyFun1 f a
  vb <- applyFun1 f b
  valueCompare name va vb
pToList = unSetPrim "to_list" $ return . ListV . S.toList
pToSet = unListPrim "to_set" $ return . SetV . S.fromList
pAny = binFoldPrim "any" $ \f xs ->
  liftM (BoolV . or) $ mapM (applyPred "any" f) xs
pAll = binFoldPrim "all" $ \f xs ->
  liftM (BoolV . and) $ mapM (applyPred "all" f) xs
pCount = binFoldPrim "count" $ \f xs ->
  liftM (intV . length) $ filterM (applyPred "count" f) xs
pMember = binAggPrim "member"
  (\v set -> return $ BoolV $ S.member v set)
  (\v list -> return $ BoolV $ elem v list)
pSize = unAggPrim "size" (return . intV . S.size) (return . intV . length)
pEmpty = unAggPrim "empty" (return . BoolV . S.null) (return . BoolV . null)
pFilter = binAggPrim "filter"
  (\x set -> do
    fn <- toFunction "filter" x
    liftM SetV $ setFilterM (applyPred "filter" fn) set)
  (\x lis -> do
    fn <- toFunction "filter" x
    liftM ListV $ filterM (applyPred "filter" fn) lis)
applyPred name fn x = applyFun1 fn x >>= toBool ("return value of function passed to " ++ name)

pUnion = liftSetBinOp "union" S.union
pUnions = unSetFoldPrim "unions" $ \ss -> return $ SetV $ S.unions ss
pIntersection = liftSetBinOp "intersection" S.intersection
pDifference = liftSetBinOp "difference" S.difference
liftSetBinOp name f = binSetPrim name $ \a b -> return $ SetV $ f a b
pInsert = binPrim "insert" $ \v x -> do
  s <- toSet' "insert" x
  return $ SetV $ S.insert v s
pDelete = binPrim "delete" $ \v x -> do
  s <- toSet' "delete" x
  return $ SetV $ S.delete v s
pImage = binPrim "image" $ \x y -> do
  fn <- toFunction "image" x
  s <- toSet' "image" y
  liftM SetV $ setMapM (applyFun1 fn) s

pDontUse = unSkillSpecPrim "dont_use" $ return . ActionV . disable
pPrioritize = unSkillSpecPrim "prioritize" $ return . ActionV . prioritize
pPrioritizeBelow = binSkillSpecPrim "prioritize_below" $ \a b -> return $ ActionV $ prioritizeBelow a b
pPrioritizeAbove = binSkillSpecPrim "prioritize_above" $ \a b -> return $ ActionV $ prioritizeAbove a b
pPrioritizeBottom = unSkillSpecPrim "prioritize_bottom" $ return . ActionV . prioritizeBottom
pUse = unSkillSpecPrim "use" $ return . ActionV . use
pTarget = binPrim' "target" $ \cd a b -> do
  skills <- toSkillSpec cd "target" a
  tgt <- makeTarget "target" b
  let m = M.fromList $ map (flip (,) tgt) $ S.toList skills
  return $ ActionV mempty{ paSkillTarget = m }
pTargetWithPriorities = binPrim' "target_with_priorities" $ \cd a x -> do
  slots <- toSkillSpec cd "target_with_priorities" a
  ts <- toList "target_with_priorities" x
  tgts <- mapM (makeTarget "target_with_priorities") ts
  m <- liftM M.fromList $ mapM (f cd tgts) $ S.toList slots
  return $ ActionV mempty{ paSkillTarget = m }
  where
    f cd tgts slot = do
      Just ss <- getSS cd slot
      possibles <- maybe (return S.empty) ($cd) $ skTargets $ ssSkill ss
      case find (not . S.null) $ map (S.intersection possibles) tgts of
        Nothing -> return $ (slot, S.empty)
        Just sth -> return (slot, sth)
pSwitchTo = unPrim' "switch_to" $ \cd a -> do
  sp <- decodeSkillPatternSpec "switch_to" cd a
  return $ ActionV mempty{ paSkillPattern = Just sp }
pReserve = binPrim' "reserve" $ \cd a b -> do
  slots <- toSkillSpec cd "target" a
  kind <- toPanelKind "target" b
  users <- liftM (S.fromList . map ssIndex . filter (elem kind . skCosts . ssSkill)) $ getSSs cd
  return $ ActionV $ disable $ S.difference users slots
pHold = unPrim' "hold" $ \cd a -> do
  inv <- makeInvocationSpec cd a
  let hold = S.fold (flip M.insert True) M.empty inv
  return $ ActionV mempty{ paHold = hold }
pSetReactionCount = binPrim' "set_reaction_count" $ \cd x y -> do
  spec <- toSkillSpec cd "set_reaction_count" x
  n <- toInt "set_reaction_count" y
  when (n < 0) $ err $ "set_reaction_count: negative count: " ++ show n
  return $ ActionV $ setReactionCount spec n
pModifyPlan = binPrim' "modify_plan" $ \cd x y -> do
  spec <- toPlanSpec cd "modify_plan" x
  fn <- toFunction "modify_plan" y
  let pm = M.fromList $ map (flip (,) [fn]) $ S.toList spec
  return $ ActionV mempty{ paPlanModify = pm }
pSeq = binPrim "seq" $ \x y -> do
  ax <- toAction "seq" x
  ay <- toAction "seq" y
  return $ ActionV $ mappend ax ay
pSequence = unListPrim "sequence" $ \vs -> liftM (ActionV . mconcat) $ mapM (toAction "sequence") vs

pPlanCode = unPlanPrim "plan_code" $ return . ExprV . planExpr

pKind = unPanelPrim "kind" $ return . PanelKindV . paKind
pPoison = unPanelPrim "poison" $ return . intV . paPoisoned
pConfuse = unPanelPrim "confuse" $ return . intV . paConfused
pIsPoisoned = unPanelPrim "is_poisoned" $ return . BoolV . (/=0) . paPoisoned
pIsConfused = unPanelPrim "is_confused" $ return . BoolV . (/=0) . paConfused
pIsUsed = unPanelPrim "is_used" $ return . BoolV . paUsed
pIsEnsnared = unPanelPrim "is_ensnared" $ return . BoolV . paEnsnared

pWeaken = unSlotPrim "weaken" $ return . intV . ssWeakened
pIsWeakened = unSlotPrim "is_weakened" $ return . BoolV . (/=0) . ssWeakened
pSlotSkill = unSlotPrim "slot_skill" $ return . SkillV . ssSkill

pExists = unPrim0 "exists" $ \x -> do
  r <- testExist x
  case r of
    Nothing -> typeError "exists" x "panel, skill slot, skill pattern or plan"
    Just bool -> return $ BoolV bool

pOwner = unPrim' "owner" $ \cd x -> do
  me <- testExist x
  case me of
    Nothing -> typeError "owner" x "panel, skill slot, skill pattern or plan"
    Just False -> err $ "owner: no such thing: " ++ show x
    Just True -> return ()
  let ret cd1 = return $ RChV (toRel cd cd1)
  case x of
    PanelV cd1 _ _ -> ret cd1
    SkillSlotV cd1 _ -> ret cd1
    PlanV cd1 _ -> ret cd1
    SkillPatternV cd1 _ -> ret cd1
    _ -> typeError "owner" x "panel, skill slot, skill pattern or plan"

bindCount name x = case x of
  SkillSlotV cd1 n -> liftM ssBound $ getSS_c name cd1 n
  PlanV cd1 n -> liftM psBound $ getPlanState_c name cd1 n
  PanelV cd1 m n -> liftM paBound $ getPanel_c name cd1 m n
  _ -> typeError name x "plan, panel or skill slot"
pBind = unPrim0 "bind" $ \x -> liftM intV $ bindCount "bind" x
pIsBound = unPrim0 "is_bound" $ \x -> liftM (BoolV . (/=0)) $ bindCount "is_bound" x

pSkillName = unSkillPrim "skill_name" $ \sk -> return $ StringV $ skName sk
pCost = unSkillPrim "cost" $ \sk -> return $ ListV $ map PanelKindV $ skCosts sk
skProp name test = unSkillPrim name $ return . BoolV . test
pPossibleTargets = binCharPrim "possible_targets" $ \cd1 x -> do
  sk <- toSkill "possible_targets" x
  tgts <- case skTargets sk of
    Nothing -> return S.empty
    Just f -> f cd1
  return $ SetV $ S.map fromSkillTarget tgts
pLevel = unSkillPrim "level" $ return . intV . skLevel
pElement = unPrim "element" $ \x -> do
  k <- toPanelKind "element" x
  newFun1 "element" $ \y -> do
    sk <- toSkill "element function" y
    return $ BoolV $ elem k $ skCosts sk

pHasSkill = binCharPrim "has_skill" $ \cd1 skill -> do
  x <- toSkill "has_skill" skill
  liftM BoolV $ hasSkill cd1 x
pSkills = unCharPrim "skills" $ \cd1 -> liftM (ListV . map (SkillV . ssSkill)) $ getSSs cd1
pReady = binCharPrim "ready" $ \cd1 skill -> do
  y <- toSkill "ready" skill
  liftM BoolV $ skillReady cd1 y
pSkillsReady = unCharPrim "skills_ready" $ \cd1 -> liftM (SetV . S.fromList . map SkillV) $ skillsReady cd1
pSkillSlots = unCharPrim "skill_slots" $ \cd1 -> do
  sss <- getSSs cd1
  return $ ListV $ map (SkillSlotV cd1 . ssIndex) sss
pPatterns = unCharPrim "patterns" $ \cd1 -> do
  sps <- getSPSs cd1
  return $ ListV $ map (SkillPatternV cd1 . spsIndex) sps
pCurrentPatternNumber = unCharPrim "current_pattern_number" $ liftM (maybeNil intV) . getCurrentSP
pCurrentPattern = unCharPrim "current_pattern" $ \cd1 -> do
  csp <- getCurrentSP cd1
  return $ maybeNil (SkillPatternV cd1) csp
pCurrentPanels = unCharPrim "current_panels" $ \cd1 -> withCurrentSP cd1 (ListV []) $ \csp -> do
  panels <- getPanels cd1
  return $ ListV $ map (PanelV cd1 csp . paIndex) panels
pCurrentPanelKinds = unCharPrim "current_panel_kinds" $ \cd1 -> do
  panels <- getPanels cd1
  return $ ListV $ map (PanelKindV . paKind) panels
pHolding = binCharPrim "holding" $ \cd1 x -> do
  sk <- toSkill "holding" x
  hss <- getHSs cd1
  return $ BoolV $ any (valid sk) hss
  where
    valid sk hs = invSkill (hsSource hs) == sk && hsEffective hs
pHeldInvocations = unCharPrim "held_invocations" $ \cd1 -> do
  hss <- getHSs cd1
  return $ SetV $ S.fromList $ map (InvocationV . hsSource) $ filter hsEffective hss
pPlans = unCharPrim "plans" $ \cd1 -> do
  pls <- getPresentPlans cd1
  return $ ListV $ map (PlanV cd1 . psIndex) pls

pCharName = unCharPrim "char_name" $ \cd -> liftM (StringV . fromMaybe "" . chName) $ char cd
pVitality = unCharPrim "vitality" $ liftM intV . getVit
pIntelligence = unCharPrim "intelligence" $ liftM intV . getInt
pReflexes = unCharPrim "reflexes" $ liftM intV . getRefl
pHP = unCharPrim "HP" $ liftM intV . getHp
pMaxHP = unCharPrim "max_HP" $ liftM (intV . vitToMaxHp) . getVit

pId = unPrim "id" return
pConst = unPrim "const" $ \x -> newFun1 "const" $ return . const x
pNot = unPrim "not" $ \x -> liftM (BoolV . not) $ toBool "not" x
pEq = unPrim "eq" $ \x -> newFun1 "eq" $ \y -> liftM BoolV $ equal x y
pEven = unIntegPrim "even" $ return . BoolV . even
pOdd = unIntegPrim "odd" $ return . BoolV . odd
pPow = binPrim0 "pow" $ \x y -> do
  a <- toInteg "pow" x
  b <- toInteg "pow" y
  when (b < 0) $ err $ "pow: negative exponent: " ++ show b
  return $ IntegerV $ a ^ b

pfn v@(FunctionV fcn) = (fcnName fcn, Value v)
val (x, y) = (x, Value y)
thunk (name, gen) = (name, Thunk th)
  where
    th = AppE (ConstE (FunctionV pf)) []
    pf = Fun name $ \_ -> gen

binRandPrim :: String -> (RndGen -> Value -> Eval Value) -> Value
binRandPrim name f = binPrim name $ \x y -> do
  gen <- toRandom name x
  f gen y

unDropPrim :: String -> (Drop -> Eval Value) -> Value
unDropPrim name f = unPrim name $ \x -> f =<< toDrop name x

unPatPrim :: String -> (CharDescr -> SkillPatState -> Game Value) -> Value
unPatPrim name f = unPrim0 name $ \x -> do
  (cd1, n) <- toSkillPattern name x
  sps <- getSPS_c name cd1 n
  f cd1 sps

unDmgPrim :: String -> (Damage -> Eval Value) -> Value
unDmgPrim name f = unPrim name $ \x -> f =<< toDmg name x

unIntfPrim :: String -> (Interference -> Eval Value) -> Value
unIntfPrim name f = unPrim name $ \x -> f =<< toIntf name x

unAttackPrim :: String -> (Attack -> Game Value) -> Value
unAttackPrim name f = unPrim0 name $ \x -> f =<< toAttack name x

charTurnPrim :: String -> (CharDescr -> Turn -> Game Value) -> Value
charTurnPrim name f = binPrim' name $ \cd x y -> do
  rch <- toRch name x
  t <- toTurn name y
  f (rel cd rch) t

unIntegPrim :: String -> (Integer -> Game Value) -> Value
unIntegPrim name f = unPrim0 name $ \x -> f =<< toInteg name x

unTurnPrim :: String -> (Turn -> Eval Value) -> Value
unTurnPrim name f = unPrim name $ \x -> f =<< toTurn name x

unInvPrim :: String -> (CharDescr -> Invocation -> Game Value) -> Value
unInvPrim name f = unPrim' name $ \cd x -> f cd =<< toInv name x

binAggPrim :: String -> (Value -> S.Set Value -> Eval Value) -> (Value -> [Value] -> Eval Value) -> Value
binAggPrim name setf listf = binPrim name $ \v x -> do
  r <- toAgg name x
  either (setf v) (listf v) r

unAggPrim :: String -> (S.Set Value -> Eval Value) -> ([Value] -> Eval Value) -> Value
unAggPrim name setf listf = unPrim name $ \x -> do
  r <- toAgg name x
  either setf listf r

unSetFoldPrim :: String -> ([S.Set Value] -> Eval Value) -> Value
unSetFoldPrim name f = unFoldPrim name $ \vs -> f =<< mapM (toSet' name) vs

binFoldPrim :: String -> (Function -> [Value] -> Eval Value) -> Value
binFoldPrim name f = binPrim name $ \x y -> do
  fun <- toFunction name x
  lst <- toList' name y
  f fun lst

unFoldPrim :: String -> ([Value] -> Eval Value) -> Value
unFoldPrim name f = unPrim name $ \x -> f =<< toList' name x

binSetPrim :: String -> (S.Set Value -> S.Set Value -> Eval Value) -> Value
binSetPrim name f = binPrim name $ \x y -> do
  sx <- toSet' name x
  sy <- toSet' name y
  f sx sy

unSetPrim :: String -> (S.Set Value -> Eval Value) -> Value
unSetPrim name f = unPrim name $ \x -> f =<< toSet name x

binListPrim :: String -> (Value -> [Value] -> Eval Value) -> Value
binListPrim name f = binPrim name $ \x y -> do
  lst <- toList name y
  f x lst

unListPrim :: String -> ([Value] -> Eval Value) -> Value
unListPrim name f = unPrim name $ \x -> f =<< toList name x

unSkillSpecPrim :: String -> (S.Set Int -> Game Value) -> Value
unSkillSpecPrim name f = unPrim' name $ \cd x -> f =<< toSkillSpec cd name x

binSkillSpecPrim :: String -> (S.Set Int -> S.Set Int -> Game Value) -> Value
binSkillSpecPrim name f = binPrim' name $ \cd x y -> do
  sx <- toSkillSpec cd name x
  sy <- toSkillSpec cd name y
  f sx sy

unPlanPrim :: String -> (Plan -> Game Value) -> Value
unPlanPrim name f = unPrim0 name $ \x -> do
  (cd1, n) <- toPlan name x
  ps <- getPlanState_c name cd1 n
  f $ psPlan ps

unPanelPrim :: String -> (Panel -> Game Value) -> Value
unPanelPrim name f = unPrim0 name $ \x -> do
  (cd1, m, n) <- toPanel name x
  pa <- getPanel_c name cd1 m n
  f pa

unSlotPrim :: String -> (SkillState -> Game Value) -> Value
unSlotPrim name f = unPrim0 name $ \x -> do
  (cd1, n) <- toSkillSlot name x
  ss <- getSS_c name cd1 n
  f ss

unSkillPrim :: String -> (Skill -> Eval Value) -> Value
unSkillPrim name f = unPrim name $ \x -> f =<< toSkill name x

unCharPrim name f = unPrim' name $ \cd x -> do
  rch <- toRch name x
  f (rel cd rch)

binCharPrim :: String -> (CharDescr -> Value -> Game Value) -> Value
binCharPrim name f = binPrim' name $ \cd x y -> do
  rch <- toRch name x
  f (rel cd rch) y

toRel :: CharDescr -> CharDescr -> RCh
toRel cd x
  | x == cd = Self
  | otherwise = Opp

toPlanSpec :: CharDescr -> String -> Value -> Game (S.Set Int)
toPlanSpec cd name v = case v of
  PlanV cd1 n
    | cd1 /= cd -> err $ name ++ ": not your plan: " ++ show v
    | otherwise -> from_int n
  IntegerV n -> case integerToInt n of
    Nothing -> nosuchplan
    Just i -> from_int i
  SetV vs -> liftM S.unions $ mapM (toPlanSpec cd name) $ S.toList vs
  ListV vs -> liftM S.unions $ mapM (toPlanSpec cd name) vs
  _ -> typeError name v "plan spec"
  where
    from_int n = do
      mps <- getPlanState cd n
      case mps of
        Nothing -> nosuchplan
        Just ps -> return $ S.singleton $ psIndex ps
    nosuchplan = err $ name ++ ": no such plan found: " ++ show v

toSkillSpec :: CharDescr -> String -> Value -> Game (S.Set Int)
toSkillSpec cd name v = case v of
  SkillSlotV cd1 n
    | cd1 == cd -> do
      mss <- getSS cd1 n
      return $ maybe S.empty (S.singleton . ssIndex) mss
    | otherwise -> err $ name ++ ": not your skill slot: " ++ show v
  SkillV sk -> liftM S.fromList $ skillSlots cd sk
  ListV vs -> liftM S.unions $ mapM (toSkillSpec cd name) vs
  SetV vs -> liftM S.unions $ mapM (toSkillSpec cd name) $ S.toList vs
  _ -> typeError name v "skill spec"

makeInvocationSpec :: CharDescr -> Value -> Game (S.Set Invocation)
makeInvocationSpec cd v = case v of
  SkillSlotV cd1 n
    | cd1 == cd -> fromSlots [n]
    | otherwise -> err $ "hold: not your skill slot: " ++ show v
  SkillV sk -> skillSlots cd sk >>= fromSlots
  InvocationV u -> return $ S.singleton u
  ListV vs -> liftM S.unions $ mapM (makeInvocationSpec cd) vs
  SetV vs -> liftM S.unions $ mapM (makeInvocationSpec cd) $ S.toList vs
  _ -> typeError "hold" v "skill invocation spec"
  where
    fromSlots sls = do
      hss <- getHSs cd
      return $ S.fromList $ map hsSource $ filter ok hss
      where
        ok hs = hsEffective hs && elem (invSlot $ hsSource hs) sls

mergeByM :: (Monad m) => (a -> a -> m Ordering) -> [a] -> [a] -> m [a]
mergeByM _ [] ys = return ys
mergeByM _ xs [] = return xs
mergeByM cmp xxs@(x:xs) yys@(y:ys) = do
  ord <- cmp x y
  case ord of
    LT -> liftM (x:) $ mergeByM cmp xs yys
    GT -> liftM (y:) $ mergeByM cmp xxs ys
    EQ -> liftM ((x:) . (y:)) $ mergeByM cmp xs ys

sortByM :: (Monad m) => (a -> a -> m Ordering) -> [a] -> m [a]
sortByM _ [] = return []
sortByM _ [x] = return [x]
sortByM cmp xs = do
  ys' <- sortByM cmp ys
  zs' <- sortByM cmp zs
  mergeByM cmp ys' zs'
  where
    (ys, zs) = splitAt (length xs `div` 2) xs

foldrM :: (Monad m) => (a -> b -> m b) -> b -> [a] -> m b
foldrM _ z [] = return z
foldrM f z (x:xs) = do
  r <- foldrM f z xs
  f x r

findM :: (Monad m) => (a -> m Bool) -> [a] -> m (Maybe a)
findM _ [] = return Nothing
findM f (x:xs) = do
  found <- f x
  if found
    then return $ Just x
    else findM f xs

maximumByM :: (Monad m) => (a -> a -> m Ordering) -> [a] -> m (Maybe a)
maximumByM _ [] = return Nothing
maximumByM comp (x:xs) = liftM Just $ foldM mx x xs
  where
    mx tmax a = do
      r <- comp tmax a
      return $ case r of
        LT -> a
        _ -> tmax

numberTurn :: Int -> Game (Maybe Turn)
numberTurn n = do
  turns <- getTurns
  return $ find ((==n) . turnNumber) $ M.elems turns

atkResult :: Attack -> Game (Maybe AttackResult)
atkResult atk = do
  results <- getCalendar
  rs <- turnSlice (atkTurn atk) results
  return $ find ((==atk) . arAttack) $ M.elems rs

maybeNil :: (t -> Value) -> Maybe t -> Value
maybeNil = maybe NilV

makeEventSetR = flip makeEventSet

makeEventSet :: Cal a -> (a -> Value) -> Value
makeEventSet cal cons = SetV $ S.fromList $ map cons $ M.elems cal

turnSlice :: (Event a) => Turn -> Cal a -> Game (Cal a)
turnSlice t cal = do
  (begin, mend) <- turnBounds t
  return $ case mend of
    Nothing -> snd $ splitBelow begin cal
    Just end -> calSlice begin end cal

calSlice :: Time -> Time -> Cal a -> Cal a
calSlice b a cal = fst $ splitBelow a $ snd $ splitBelow b cal

splitBelow :: (Ord k) => k -> M.Map k a -> (M.Map k a, M.Map k a)
splitBelow t m = (low, high')
  where
    (low, mid, high) = M.splitLookup t m
    high' = case mid of
      Just v -> M.insert t v high
      _ -> high

turnBoundsE :: Turn -> Eval (Time, Time)
turnBoundsE t = liftM ((,) (turnTime t)) $ turnEndTimeE t

turnEndTimeE :: Turn -> Eval Time
turnEndTimeE t = do
  now <- getNow
  liftM (fromMaybe now) $ liftGame $ turnEndTime t

turnBounds :: Turn -> Game (Time, Maybe Time)
turnBounds t = liftM ((,) (turnTime t)) $ turnEndTime t

turnEndTime :: Turn -> Game (Maybe Time)
turnEndTime t = do
  turns <- getTurns
  let (_, sub) = M.split (turnTime t) turns
  if M.null sub
    then return Nothing -- t is the current turn
    else return $ Just $ turnTime $ snd $ M.findMin sub

setMapM :: (Monad m, Ord a, Ord b) => (a -> m b) -> S.Set a -> m (S.Set b)
setMapM f set = Fold.foldl f' (return S.empty) set
  where
    f' acc v = do
      w <- f v
      z <- acc
      return $ S.insert w z

setFilterM :: (Monad m, Ord a) => (a -> m Bool) -> S.Set a -> m (S.Set a)
setFilterM f set = Fold.foldl f' (return S.empty) set
  where
    f' acc v = do
      w <- f v
      z <- acc
      return $ if w then S.insert v z else z

disable :: SkillSpec -> PlanAction
disable spec = makePriority name f
  where
    name = "disable(" ++ showSkillSpec spec ++ ")"
    f proj = filter (not . (`S.member` spec) . proj . fst)

prioritize, prioritizeBottom :: SkillSpec -> PlanAction
prioritize spec = makePriority name f
  where
    name = "prioritize(" ++ showSkillSpec spec ++ ")"
    f proj xs = tgts ++ others
      where
        (tgts, others) = partition ((`S.member` spec) . proj . fst) xs
prioritizeBottom spec = makePriority name f
  where
    name = "prioritize_bottom(" ++ showSkillSpec spec ++ ")"
    f proj xs = others ++ tgts
      where
        (tgts, others) = partition ((`S.member` spec) . proj . fst) xs

prioritizeAbove, prioritizeBelow :: SkillSpec -> SkillSpec -> PlanAction
prioritizeAbove spec rel = makePriority name f
  where
    name = "prioritize_above(" ++ showSkillSpec spec ++ "|" ++ showSkillSpec rel ++ ")"
    f proj xs = shift spec rel proj xs
prioritizeBelow spec rel = makePriority name f
  where
    name = "prioritize_below(" ++ showSkillSpec spec ++ "|" ++ showSkillSpec rel ++ ")"
    f proj xs = reverse $ shift spec rel proj $ reverse xs

shift :: SkillSpec -> SkillSpec -> (a -> Int) -> [(a, Bool)] -> [(a, Bool)]
shift spec rel proj xs = pre ++ tgt ++ post
  where
    (pre, post) = break ((`S.member` rel) . proj . fst) others
    (tgt, others) = partition ((`S.member` spec) . proj . fst) xs

setReactionCount :: SkillSpec -> Int -> PlanAction
setReactionCount spec n = makePriority name f
  where
    name = "reaction_count(" ++ showSkillSpec spec ++ "|" ++ show n ++ ")"
    f proj = concatMap g
      where
        g m@(x, _)
          | S.member (proj x) spec = replicate n m
          | otherwise = [m]

use :: SkillSpec -> PlanAction
use spec = makePriority name f
  where
    name = "use(" ++ showSkillSpec spec ++ ")"
    f proj = map g
      where
        g m@(x, _)
          | S.member (proj x) spec = (x, True)
          | otherwise = m

showSkillSpec :: SkillSpec -> String
showSkillSpec ss = concat $ intersperse "," $ map show $ S.toList ss

makePriority :: String -> (forall a. (a -> Int) -> [(a, Bool)] -> [(a, Bool)]) -> PlanAction
makePriority name f = mempty{ paSkillPriority = [SPM name f] }

type SkillSpec = S.Set Int

skillVars :: [(Name, Entity)]
skillVars = map f allSkills
  where
    f sk = (skName sk, Value $ SkillV sk)

panelKinds :: [(Name, Entity)]
panelKinds = map f [minBound .. maxBound]
  where
    f k = (showKind k, Value $ PanelKindV k)

makeTarget :: String -> Value -> Game (S.Set SkillTarget)
makeTarget name x = case x of
  SkillSlotV cd1 n -> return $ S.singleton $ SkillTarget cd1 n
  SkillV sk -> liftM S.fromList $ liftM2 (++) (slots First) (slots Second)
    where
      slots cd1 = liftM (map (SkillTarget cd1)) $ skillSlots cd1 sk
  PanelV cd1 m n -> return $ S.singleton $ PanelTarget cd1 m n
  PanelKindV k -> liftM S.fromList $ liftM2 (++) (panels First) (panels Second)
    where
      panels cd1 = liftM (map (uncurry $ PanelTarget cd1)) $ kindPanels cd1 k
  PlanV cd1 n -> return $ S.singleton $ PlanTarget cd1 n
  ListV vs -> liftM S.unions $ mapM (makeTarget name) vs
  SetV vs -> liftM S.unions $ mapM (makeTarget name) $ S.toList vs
  _ -> typeError name x "skill target"

decodeSkillPatternSpec :: String -> CharDescr -> Value -> Game Int
decodeSkillPatternSpec loc cd x = case x of
  SkillPatternV cd1 n
    | cd1 == cd -> return n
    | otherwise -> err $ loc ++ ": not your skill pattern: " ++ show x
  IntegerV ni -> case integerToInt ni of
    Nothing -> err $ loc ++ ": invalid index: " ++ show ni
    Just n -> do
      msp <- getSPS cd n
      when (isNothing msp) $ err $ loc ++ ": invalid index: " ++ show n
      return $ fromIntegral n
  StringV name -> do
    msp <- getSPSByName cd name
    case msp of
      Just sps -> return $ spsIndex sps
      Nothing -> err $ loc ++ ": skill pattern not found: " ++ name
  _ -> typeError loc x "skill pattern spec"

getCurrentFirstInRel :: Eval RCh
getCurrentFirstInRel = do
  cd <- getSelf
  liftM (toRel cd) $ liftGame getFirstToMove

allEvents :: (Event a) => (a -> Value) -> Eval Value
allEvents cons = liftM (makeEventSetR cons) $ liftGame getCalendar

turnEvents :: (Event a) => String -> (a -> Value) -> (a -> CharDescr) -> Value
turnEvents name cons chr = charTurnPrim name $ \cd1 t -> do
  evts <- getCalendar
  sl <- liftM (M.filter ((==cd1) . chr)) $ turnSlice t evts
  return $ makeEventSet sl cons

packToString :: [Value] -> Eval String
packToString vs = mapM f vs
  where
    f v = do
      i <- toInt "pack" v
      if i <= 0 || 0x10ffff < i
        then err $ "codepoint out of range: " ++ show i
        else return $ chr i

getSS_c :: String -> CharDescr -> Int -> Game SkillState
getSS_c loc cd n = do
  mss <- getSS cd n
  case mss of
    Nothing -> err $ loc ++ ": no such slot: " ++ show n
    Just ss -> return ss

getPlanState_c :: String -> CharDescr -> Int -> Game PlanState
getPlanState_c loc cd n = do
  mps <- getPlanState cd n
  case mps of
    Nothing -> err $ loc ++ ": no such plan: " ++ show n
    Just ps -> return ps

getPanel_c :: String -> CharDescr -> Int -> Int -> Game Panel
getPanel_c loc cd m n = do
  mpa <- getPanel cd m n
  case mpa of
    Nothing -> err $ loc ++ ": no such panel: " ++ show (m, n)
    Just pa -> return pa

getSPS_c :: String -> CharDescr -> Int -> Game SkillPatState
getSPS_c loc cd n = do
  msps <- getSPS cd n
  case msps of
    Nothing -> err $ loc ++ ": no such pattern: " ++ show n
    Just sps -> return sps

testExist :: Value -> Game (Maybe Bool)
testExist v = case v of
  PanelV cd1 m n -> ret $ getPanel cd1 m n
  SkillSlotV cd1 n -> ret $ getSS cd1 n
  PlanV cd1 n -> ret $ getPlanState cd1 n
  SkillPatternV cd1 n -> ret $ getSPS cd1 n
  _ -> return Nothing
  where
    ret = liftM (Just . isJust)
