module Match
  ( match
  , quietMatch
  , matchEx
  , newGameInfo
  , newMatch
  , replayMatch
  , newMatch_
  , Option(..)
  , defaultOption
  , sscOut
  , RndGen
  ) where

import Game
import Run
import GameUtils
import CommonOperations

import Data.Dynamic
import Control.Monad.Reader
import Control.Exception
import System.Random
import Data.Maybe
import Data.List
import System.IO.SSC.Haskell98 as SSC

gameOvers :: Exception -> Maybe (GameResult, GameInfo)
gameOvers exn = do
  v <- dynExceptions exn
  GFE gr gi <- fromDynamic v
  return (gr, gi)

match, quietMatch :: Character -> Character -> IO GameResult
match = newMatch_ (return ()) True
quietMatch = newMatch_ (return ()) False

replayMatch :: RndGen -> Character -> Character -> IO GameResult
replayMatch g x y = liftM fst $ matchEx opt x y
  where
    opt = defaultOption{ optGen = Just g, optLogOut = sscOut }

newMatch_ :: Game () -> Bool -> Character -> Character -> IO GameResult
newMatch_ hook verbose x y = liftM fst $ newMatch hook verbose x y

sscOut :: Maybe (String -> IO ())
sscOut = Just SSC.putStrLn

newMatch :: Game () -> Bool -> Character -> Character -> IO (GameResult, RndGen)
newMatch hook verbose x y = do
  g <- newStdGen
  let opt = defaultOption{ optGen = Just g, optHook = hook, optLogOut = if verbose then sscOut else Nothing }
  r <- liftM fst $ matchEx opt x y
  return (r, g)

data Option = Opt
  { optGen :: Maybe RndGen
  , optLogOut :: Maybe (String -> IO ())
  , optHook :: Game ()
  , optNGame :: Int
  }

defaultOption :: Option
defaultOption = Opt
  { optGen = Nothing
  , optLogOut = Nothing
  , optHook = return ()
  , optNGame = 3
  }

matchEx :: Option -> Character -> Character -> IO (GameResult, GameInfo)
matchEx option x y = catchJust gameOvers go return
  where
    go = do
      gen <- maybe newStdGen return $ optGen option
      foldM exec gen $ replicate ngame (game, False) ++ [(game, True), (firstWins, False)]
      fail "impossible!"

    ngame = optNGame option

    exec gen (action, sd) = do
      gi <- newGameInfo sd gen (optLogOut option) (optHook option) x y
      runGame action' gi
      where
        action' = do
          action
          getRndGen

firstWins :: Game ()
firstWins = do
  (a, _) <- moveOrder
  win a

game :: Game ()
game = do
  mapM_ turn [0..29]
  mes "30ターン経過のためゲーム終了"
  where
    turn n = do
      (a, b) <- moveOrder
      startTurn n a
      printTurnInfo
      skillPatternPhase
      statusRecoverPhase
      statusEffectPhase
      action a
      action b
      endPhase

moveOrder :: Game (CharDescr, CharDescr)
moveOrder = do
  va <- roll First
  vb <- roll Second
  case compare va vb of
    LT -> do
      --mes "二者先攻"
      return (Second, First)
    GT -> do
      --mes "一者先攻"
      return (First, Second)
    EQ -> moveOrder
  where
    roll cd = do
      vit <- getVit cd
      m <- dice
      return $ vit + m

startTurn :: Int -> CharDescr -> Game ()
startTurn n first = do
  (now, _) <- tick
  let
    turn = Turn
      { turnTime = now
      , turnNumber = n
      , turnFirstToMove = first
      }
  setCurrentTurn turn
  recordTurn turn

skillPatternPhase :: Game ()
skillPatternPhase = do
  mf <- askSkillPattern First
  ms <- askSkillPattern Second
  setSkillPattern First mf
  setSkillPattern Second ms

setSkillPattern :: CharDescr -> (Maybe Int) -> Game ()
setSkillPattern cd nsp = withCurrentSP cd () $ \csp -> case nsp of
  Just sp
    | sp /= csp -> changeSkillPattern cd csp sp
  _ -> return ()

changeSkillPattern :: CharDescr -> Int -> Int -> Game ()
changeSkillPattern cd from to = do
  mesC cd $ "はスキルパターンを" ++ show to ++ "に変更した"
  setCurrentSP cd to 
  hss <- getHSs cd
  let ins = filter (belong to) hss
  let outs = filter (belong from) hss
  mapM_ (stopHeldSkillByUnique cd . hsUnique) outs
  mapM_ (restartHeldSkillByUnique cd . hsUnique) ins
  where
    belong sp hs = any ((==sp) . fst) $ hsCosts hs

stopHeldSkillByUnique :: CharDescr -> Unique -> Game ()
stopHeldSkillByUnique cd un = do
  hss <- getHSs cd
  Just idx <- return $ findIndex ((==un) . hsUnique) hss
  mesC cd $ "の維持している" ++ show (hss !! idx) ++ "が効力を失った"
  hsStop (hss!!idx)
  modifyHS cd idx $ \hs -> hs{ hsEffective = False }

restartHeldSkillByUnique :: CharDescr -> Unique -> Game ()
restartHeldSkillByUnique cd un = do
  hss <- getHSs cd
  Just idx <- return $ findIndex ((==un) . hsUnique) hss
  mesC cd $ "の維持している" ++ show (hss !! idx) ++ "が復帰した"
  hsStart (hss!!idx)
  modifyHS cd idx $ \hs -> hs{ hsEffective = True }

statusRecoverPhase :: Game ()
statusRecoverPhase = recov First >> recov Second
  where
    recov cd = do
      recoverPlans cd
      recoverSkills cd
      recoverPanels cd

recoverPlans :: CharDescr -> Game ()
recoverPlans cd = do
  pss <- getPresentPlans cd
  fs <- recov pss
  mapM_ (curePlanBind cd) fs
  where
    recov pss = liftM concat $ mapM g pss
    g PS{ psBound = bound, psIndex = n } = do
      f <- fixes bound
      return $ replicate f n

recoverSkills :: CharDescr -> Game ()
recoverSkills cd = do
  sss <- getSSs cd
  idxs <- liftM concat $ mapM binds sss
  mapM_ (cureSkillBind cd) idxs
  where
    binds SS{ ssBound = bound, ssIndex = n } = do
      f <- fixes bound
      return $ replicate f n

recoverPanels :: CharDescr -> Game ()
recoverPanels cd = withCurrentSP cd () $ \csp -> do
  panels <- getPanels cd
  (confs, binds) <- liftM unzip $ mapM confs_and_binds panels
  mapM_ (cureConfuse cd csp) $ concat confs
  mapM_ (curePanelBind cd csp) $ concat binds
  where
    confs_and_binds Pa{ paConfused = conf, paBound = bind, paIndex = n } = do
      cf <- fixes conf
      bf <- fixes bind
      return $ (replicate cf n, replicate bf n)

fixes :: Int -> Game Int
fixes n = loop 0 0
  where
    loop i s
      | i == n = return s
      | otherwise = do
        v <- rnd autoCureProbDenom
        loop (i+1) (if v < autoCureProbNum then s + 1 else s)

statusEffectPhase :: Game ()
statusEffectPhase = do
  eps <- paradiseLostEffect
  efs <- effect First
  ess <- effect Second
  performDelays $ eps ++ efs ++ ess
  where
    effect cd = confuseEffect cd

paradiseLostEffect :: Game [Delay]
paradiseLostEffect = do
  pl <- getParadiseLost
  liftM concat $ replicateM pl $ do
    mes $ "崩れゆく聖域が効果中"
    liftM2 (++) (ruin First) (ruin Second)
  where
    ruin cd = do
      vit <- getVit cd
      refl <- getRefl cd
      int <- getInt cd
      let 
        p 0 _ = Nothing
        p _ x = Just x
      r <- chooseMaybe $ catMaybes [p vit 0, p refl 1, p int 2]
      case r of
        Just 0 -> loseVit cd vit
        Just 1 -> loseRefl cd refl
        Just 2 -> loseInt cd int
        Nothing -> loseNothing cd

    loseVit cd _ = do
      decrVit cd 1
      return []

    loseRefl cd refl = do
      mesC cd $ "は神経反射を1点失った"
      setRefl cd (refl - 1)
      spss <- getSPSs cd
      lps <- liftM catMaybes $ mapM choosePanel spss
      liftM concat $ mapM (uncurry $ losePanel cd) lps
      where
        choosePanel SPS{ spsPanels = panels, spsIndex = idx }
          | npa == refl = do
              k <- liftM paIndex $ choose panels
              return $ Just (idx, k)
          | otherwise = return Nothing
          where
            npa = length panels

    loseInt cd int = do
      mesC cd $ "は知力を1点失った"
      setInt cd (int - 1)
      plan
      liftM2 (++) slot pattern
      where
        plan = do
          plans <- getPresentPlans cd
          when (length plans == int) $ do
            z <- liftM psIndex $ choose plans
            losePlan cd z
        slot = do
          slots <- getSSs cd
          if (length slots /= int + nFreeSkill) then return [] else do
            z <- liftM ssIndex $ choose slots
            loseSlot cd z
            return []
        pattern = do
          pats <- getSPSs cd
          if length pats /= int then return [] else do
            pat <- choose pats
            let z = spsIndex pat
            mesC cd $ "はスキルパターン" ++ show z ++ "を失った"
            oldsp <- getCurrentSP' cd
            deleteSP cd z
            when (oldsp == z && int >= 2) $ do -- FIXME: don't repeat yourself!
              newsp <- liftM spsIndex $ choose =<< getSPSs cd
              mesC cd $ "のスキルパターンが" ++ show newsp ++ "に変更された"
              setCurrentSP cd newsp
              hss <- getHSs cd
              mapM_ (restartHeldSkillByUnique cd . hsUnique) $ filter (any ((==newsp) . fst) . hsCosts) hss
            liftM concat $ mapM (uncurry $ notifyPanelChange cd DropByLostPanel) $ map ((,) z . paIndex) $ spsPanels pat

    loseNothing cd = do
      mesC cd $ "に失うものは何もない"
      return []

confuseEffect :: CharDescr -> Game [Delay]
confuseEffect cd = withCurrentSP cd [] $ \csp -> do
  panels <- getPanels cd
  liftM concat $ mapM (rollPanel cd csp) $ concatMap confs panels
  where
    confs pa = replicate (paConfused pa) (paIndex pa)

endPhase :: Game ()
endPhase = do
  hf <- askHoldOrNot First
  hs <- askHoldOrNot Second
  (df, toContF) <- releasePanels First hf
  (ds, toContS) <- releasePanels Second hs
  mapM_ continueSkill $ sortOn hsUnique $ toContF ++ toContS
  performDelays $ df ++ ds

releasePanels :: CharDescr -> (Invocation -> Bool) -> Game ([Delay], [HeldSkill])
releasePanels cd hold = withCurrentSP cd ([], []) $ \csp -> do
  rs <- unholdSkills csp
  used <- panelsUsed
  panels <- getPanels cd
  modifyCurrentPanels cd $ concatMap (ops csp used) panels
  return rs
  where
    panelsUsed = do
      hss <- getHSs cd
      return $ concatMap hsCosts hss
    ops csp used pa
      | paEnsnared pa = [(idx, SetEnsnared False)]
      | elem (csp, idx) used  = []
      | otherwise = [(idx, SetUsed False)]
      where
        idx = paIndex pa
    unholdSkills csp = do
      held <- getHSs cd
      ps <- getPanels cd
      let bad (m, n) = m /= csp || paEnsnared (fromJust $ find ((==n) . paIndex) ps)
      let toHold hs = any bad (hsCosts hs) || hold (hsSource hs)
      let (held', todrop) = partition toHold held
      setHSs cd held'
      delays <- liftM concat $ mapM (terminateHeldSkill DropByWill) todrop
      return (delays, held')

continueSkill :: HeldSkill -> Game ()
continueSkill hs = do
  (now, turn) <- tick
  recordEvent Cont{ contTime = now, contTurn = turn, contInv = hsSource hs }

printTurnInfo :: Game ()
printTurnInfo = do
  n <- getTurnCount
  f <- getFirstToMove >>= getPrettyName
  mes ""
  mes $ "=== (Turn " ++ show n ++ ", 先攻: " ++ f ++ ")"
  printCharInfo First
  printCharInfo Second
  mes "==="

action :: CharDescr -> Game ()
action cd = do
  sds <- plannedSkills cd (isJust . skEffect) Nothing Nothing
  try_skills sds
  where
    try_skills [] = return ()
    try_skills ((sd@(SD _ _ sk), forcep):rest)
      | isNothing (skEffect sk) = try_skills rest
      | otherwise = do
        cond <- skCondition sk
        if evalSkillCondition forcep cond
          then do
            activeInvoke sd
            action cd
          else try_skills rest

dice :: Game Int
dice = do
  v <- liftM (1+) $ rnd 6
  --mes $ "さいころ: " ++ show v
  return v

printCharInfo :: CharDescr -> Game ()
printCharInfo cd = do
  out <- isOutputPresent
  when out $ do
    s <- formatCharInfo cd
    mapM_ mes $ lines s

formatCharInfo :: CharDescr -> Game String
formatCharInfo cd = do
  name <- getPrettyName cd
  curhp <- getHp cd
  vit <- getVit cd
  refl <- getRefl cd
  int <- getInt cd
  sss <- getSSs cd
  pss <- getPresentPlans cd
  hss <- getHSs cd
  spss <- getSPSs cd
  mcsp <- getCurrentSP cd
  let
    maxhp = vitToMaxHp vit
    statusLine = unwords 
      [ name
      , "HP:" ++ show curhp ++ "/" ++ show maxhp
      , "VIT:" ++ show vit
      , "REF:" ++ show refl
      , "INT:" ++ show int
      ]
    skillPatternLine = unwords $ map f spss
      where
        f sps = "(" ++ spsName sps ++ ":" ++ concatMap g (spsPanels sps) ++ ")" ++ if Just (spsIndex sps) == mcsp then "*" else ""
        g pa = showKind (paKind pa) ++ bracket addition
          where
            addition = used ++ conf ++ poison ++ bind ++ ensn
            used = if paUsed pa then "済" else ""
            conf = many "乱" (paConfused pa)
            poison = many "毒" (paPoisoned pa)
            bind = many "縛" (paBound pa)
            ensn = if paEnsnared pa then "陥" else ""
    many _ 0 = ""
    many str 1 = str
    many str n = str ++ show n
    bracket "" = ""
    bracket xs = "[" ++ xs ++ "]"
    skillLine = unwords $ map f sss
      where
        f ss = show (ssSkill ss) ++ bracket (many "縛" (ssBound ss) ++ many "弱" (ssWeakened ss))
    heldLine = if null hl then Nothing else Just hl
    hl = unwords $ catMaybes [holds, boundPlans]
    holds = list "維持" $ map f hss
      where
        f hs = show (invSkill $ hsSource hs) ++ if hsEffective hs then "" else "[休]"
    boundPlans = list "束縛プラン" $ mapMaybe f pss
      where
        f ps = manyb (show $ psIndex ps) (psBound ps)
    manyb _ 0 = Nothing
    manyb str 1 = Just str
    manyb str n = Just $ str ++ "[" ++ show n ++ "]"
    list _ [] = Nothing
    list title xs = Just $ title ++ ": " ++ concat (intersperse "," xs)

  return $ unlines $ catMaybes 
    [ Just $ statusLine ++ " | " ++ skillPatternLine
    , Just skillLine
    , heldLine
    ] 

