module Parser
  ( charsFromFile
  , parseChars
  , parsePlan
  , parseSkill
  , parsePattern
  , parseExp
  , parseExpE
  , exprP
  , planP
  , skillP
  , patternP
  , languageDef
  ) where

import Types
import Plan
import ExprUtils
import CharBuilder
import AllSkills

import Text.ParserCombinators.Parsec as P
import Text.ParserCombinators.Parsec.Language as P
import Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Expr as P
import Control.Monad
import qualified Data.Set as S
import Data.List
import System.IO.SSC.Haskell98 as SSC
import Control.Monad.Error
import Data.Maybe
import Data.Monoid(mempty)

type Par = GenParser Char ()

tp :: TokenParser ()
tp = makeTokenParser languageDef

languageDef :: LanguageDef ()
languageDef = haskellStyle
  { reservedOpNames = 
      ["!", "~", "*", "/", "%", "+", "-", "==", "/=", "<", ">", "<=", ">=",
       "&&", "||", "&&&", "|||", "?", "=", ":", "->", "\\", "=>", "|", "#",
       ":=", ".", "&", "'"]
  , identStart = add <|> identStart haskellStyle
  , identLetter = add <|> identLetter haskellStyle
  }
  where
    add = oneOf "＝_"

exprP :: Par Expr
exprP = do
  left <- exprNoThenP
  (<|> return left) $ do
    reservedOp tp "=>"
    right <- exprNoThenP
    return $ IfE left right $ ConstE $ ActionV mempty

exprNoThenP :: Par Expr
exprNoThenP = lambdaP <|> nonLambdaP <?> "expression"

lambdaP :: Par Expr
lambdaP = (<?> "lambda expression") $ do
  reservedOp tp "\\"
  vars <- many1 $ identifier tp
  reservedOp tp "->"
  body <- exprNoThenP
  return $ AbsE vars body

nonLambdaP :: Par Expr
nonLambdaP = do
  left <- binExprP
  ifTailP left <|> letTailP left <|> return left

letTailP :: Expr -> Par Expr
letTailP (RefE var) = do
  cons <- consP
  init <- exprP
  reservedOp tp "|"
  body <- exprNoThenP
  return $ cons var init body
  where
    consP = a <|> b
    a = reservedOp tp "=" >> return LetE
    b = reservedOp tp ":=" >> return LetNameE
letTailP _ = mzero

ifTailP :: Expr -> Par Expr
ifTailP cond = do
  reservedOp tp "?"
  tb <- exprP
  reservedOp tp "|"
  fb <- exprNoThenP
  return $ IfE cond tb fb

binExprP :: Par Expr
binExprP = buildExpressionParser tbl appP
  where
    tbl =
      [ [Prefix negP]
      , [Infix atP AssocLeft]
      , [Infix compP AssocRight]
      , [Infix fandP AssocLeft]
      , [Infix forP AssocLeft]
      , [Infix (arith "*" (*)) AssocLeft, Infix (arithz "/" div) AssocLeft, Infix (arithz "%" mod) AssocLeft]
      , [Infix (arith "+" (+)) AssocLeft, Infix (arith "-" (-)) AssocLeft]
      , [Infix eqP AssocNone, Infix neqP AssocNone] ++
         map (\(name, flip, neg) -> Infix (rel name flip neg) AssocNone)
          [("<", False, False), (">", True, False), ("<=", True, True), (">=", False, True)]
      , [Infix andP AssocLeft]
      , [Infix orP AssocLeft]
      , [Infix timeShiftP AssocNone] 
      ]
    opr name val = reservedOp tp name >> return val
    opr1 name exp = opr name $ \a -> AppE exp [a]
    opr2 name exp = opr name $ \a b -> AppE exp [a, b]
    arith name op = opr2 name $ liftArith name op
    arithz name op = opr2 name $ liftArithZ name op
    rel name flip neg = opr2 name $ compareValue name flip neg
    andP = opr "&&" $ \a b -> IfE a b (ConstE $ BoolV False)
    orP = opr "||" $ \a b -> IfE a (ConstE $ BoolV True) b
    negP = opr1 "~" $ unPrimE "~" $ \x -> liftM (IntegerV . negate) $ toInteg "~" x
    eqP = opr2 "==" $ binPrimE "==" $ \x y -> liftM BoolV $ equal x y
    neqP = opr2 "/=" $ binPrimE "/=" $ \x y -> liftM (BoolV . not) $ equal x y
    atP = opr2 "!" $ binPrimE "!" $ \x y -> at x y
    compP = opr2 "." $ binFunPrimE "." compose
    timeShiftP = opr "#" $ \a b -> TimeShiftE a b
    fandP = opr2 "&&&" $ binFunPrimE "&&&" fAnd
    forP = opr2 "|||" $ binFunPrimE "|||" fOr

fAnd :: Function -> Function -> Value -> Eval Value
fAnd f g x = do
  c <- toBool "&&&" =<< applyFun1 f x
  if c then applyFun1 g x else return $ BoolV False

fOr :: Function -> Function -> Value -> Eval Value
fOr f g x = do
  c <- toBool "|||" =<< applyFun1 f x
  if c then return $ BoolV True else applyFun1 g x

compose :: Function -> Function -> Value -> Eval Value
compose g f x = applyFun1 f x >>= applyFun1 g

binFunPrimE :: String -> (Function -> Function -> Value -> Eval Value) -> Expr
binFunPrimE name work = binPrimE name $ \x y -> do
  f <- toFunction name x
  g <- toFunction name y
  newFun1 name $ \v -> work f g v

at :: Value -> Value -> Eval Value
at lst idx = do
  vs <- toList "!" lst
  n <- toInt "!" idx
  if n >= length vs
    then err $ "!: index too large"
    else if n < 0
      then err $ "!: negative index"
      else return $ vs !! n

appP :: Par Expr
appP = do
  leftfn <- quoteP
  argss <- many $ do
    args <- exprNoThenP `sepBy1` comma tp
    optional $ semi tp
    return args
  return $ foldl AppE leftfn argss

quoteP :: Par Expr
quoteP = (<|>primP) $ do
  reservedOp tp "'"
  e <- quoteP
  return $ ConstE $ ExprV e

primP :: Par Expr
primP = foldr1 (<|>) 
  [ (liftM (ConstE . IntegerV) $ natural tp) <?> "integer literal"
  , (liftM (ConstE . StringV) $ stringLiteral tp) <?> "string literal"
  , (liftM RefE $ identifier tp) <?> "identifier"
  , parens tp exprP
  , setP
  , listP
  ]

setP :: Par Expr
setP = (<?> "set expression") $ braces tp $ do
  es <- exprP `sepBy` reservedOp tp ":"
  return $ AppE (ConstE pConsSet) es

listP :: Par Expr
listP = (<?> "list expression") $ brackets tp $ do
  es <- exprP `sepBy` reservedOp tp ":"
  return $ AppE (ConstE pConsList) es

planP :: Par Plan
planP = do
  exp <- exprP
  return $ Plan (Code exp Nothing)

patternP :: Par (Maybe String, SkillPattern)
patternP = do
  mnm <- option Nothing $ do
    nm <- stringLiteral tp
    colon tp
    return $ Just nm
  idt <- identifier tp
  pat <- mapM f idt
  return (mnm, pat)
  where
    f c = case c of
      '地' -> return Earth
      '火' -> return Fire
      '水' -> return Water
      '風' -> return Wind
      '剣' -> return Sword
      _ -> fail $ "patternP: not a panel kind: " ++ [c]

skillP :: Par Skill
skillP = do
  idt <- identifier tp
  case find ((==idt) . skName) allSkills of
    Just sk -> return sk
    Nothing -> fail $ "skill not found: " ++ idt

data PreChar = PC
  { pcName :: Maybe String
  , pcParams :: Maybe (Int, Int, Int)
  , pcPatterns :: Maybe [(Maybe String, SkillPattern)]
  , pcSkills :: Maybe [Skill]
  , pcPlans :: Maybe [Plan]
  }
type AttrReader = (String, Par (PreChar -> Maybe PreChar))

preCharToChar :: PreChar -> Either String Character
preCharToChar pc = do
  name <- frm "name" $ pcName pc
  params <- frm "params" $ pcParams pc
  pats <- frm "patterns" $ pcPatterns pc
  skills <- frm "skills" $ pcSkills pc
  plans <- frm "plans" $ pcPlans pc
  return $ buildCharacter (Just name) params pats skills plans
  where
    frm tag v = case v of
      Nothing -> fail $ "field not specified: " ++ tag
      Just x -> return x

prop :: String -> Par a -> (PreChar -> Maybe a) -> (Maybe a -> PreChar -> PreChar) -> AttrReader
prop pname parser getter setter = (pname, pp)
  where
    pp = do
      v <- parser
      return $ \pc -> case getter pc of
        Just _ -> Nothing
        Nothing -> Just $ setter (Just v) pc

nameP, paramsP, patternsP, skillsP, plansP :: AttrReader
nameP = prop "name" (stringLiteral tp) pcName $ \nm pc -> pc{ pcName = nm }
paramsP = prop "params" p pcParams $ \v pc -> pc{ pcParams = v }
  where
    p = do
      ps <- blist $ liftM fromIntegral $ integer tp
      case ps of
        [v, r, i] -> return (v, r, i)
        _ -> fail $ "parameter list must be a triple: " ++ show ps
patternsP = prop "patterns" (blist patternP) pcPatterns $ \v pc -> pc{ pcPatterns = v }
skillsP = prop "skills" (blist skillP) pcSkills $ \v pc -> pc{ pcSkills = v }
plansP = prop "plans" (balist planP) pcPlans $ \v pc -> pc{ pcPlans = v }

attrs :: [AttrReader]
attrs = [nameP, paramsP, patternsP, skillsP, plansP]

attrP :: Par (PreChar -> Either String PreChar)
attrP = do
  idt <- identifier tp
  reservedOp tp "="
  case lookup idt attrs of
    Just m -> do
      f <- m
      return $ \pc -> case f pc of
        Nothing -> fail $ "duplicate field: " ++ idt
        Just x -> return x
    Nothing -> fail $ "unknown field: " ++ idt

charP :: Par Character
charP = do
  attrs <- blist attrP
  prec <- liftError $ foldM (flip ($)) initPrec attrs
  liftError $ preCharToChar prec

liftError :: Either String a -> Par a
liftError = either fail return

initPrec :: PreChar
initPrec = PC
  { pcName = Nothing
  , pcParams = Nothing
  , pcPatterns = Nothing
  , pcSkills = Nothing
  , pcPlans = Nothing
  }

blist, balist :: Par a -> Par [a]
blist p = braces tp $ p `sepBy` comma tp
balist p = braces tp $ p `sepBy` reservedOp tp "&"

parseExp :: String -> Expr
parseExp input = either error id $ parseExpE input

wholeInput :: Par a -> Par a
wholeInput p = do
  whiteSpace tp
  e <- p
  eof
  return e

standalone :: Par a -> String -> Either String a
standalone p input = case parse (wholeInput p) "" input of
  Left err -> Left $ show err
  Right v -> Right v

parseExpE :: String -> Either String Expr
parseExpE = standalone exprP

parseChars :: String -> Either String [Character]
parseChars = standalone $ many charP

parsePlan :: String -> Either String Plan
parsePlan = standalone planP

parsePattern :: String -> Either String (Maybe String, SkillPattern)
parsePattern = standalone patternP

parseSkill :: String -> Either String Skill
parseSkill = standalone skillP

charsFromFile :: String -> IO [Character]
charsFromFile fname = do
  s <- SSC.readFile fname
  chs <- either fail return $ parseChars s
  return chs
  mapM cc chs
  where
    cc ch = do
      mapM_ rep msgs
      return ch'
      where
        (msgs, ch') = compileChar ch
        rep msg = SSC.putStrLn $ "compiler warning: " ++ (fromJust $ chName ch) ++ ": " ++ msg

pConsSet :: Value
pConsSet = FunctionV $ Fun "cons-set" $ \vs -> return $ SetV $ S.fromList vs

pConsList :: Value
pConsList = FunctionV $ Fun "cons-list" $ \vs -> return $ ListV vs

liftArith :: String -> (Integer -> Integer -> Integer) -> Expr
liftArith name op = binIntPrim name $ \x y -> return $ IntegerV $ op x y

liftArithZ :: String -> (Integer -> Integer -> Integer) -> Expr
liftArithZ name op = binIntPrim name $ \x y -> case y of
  0 -> err $ name ++ ": division by 0"
  _ -> return $ IntegerV $ op x y

compareValue :: String -> Bool -> Bool -> Expr
compareValue name flip neg = binPrimE name $ \x y -> do
  r <- if flip then f y x else f x y
  return $ BoolV $ (if neg then not else id) $ r == LT
  where
    f = valueCompare name

liftRel :: String -> (Integer -> Integer -> Bool) -> Expr
liftRel name op = binIntPrim name $ \x y -> return $ BoolV $ op x y

