{-# LANGUAGE Rank2Types #-}
module Plan where

import Game

import Text.PrettyPrint.HughesPJ
import qualified Data.Set as S
import qualified Data.Map as M
import Data.Monoid(Monoid(..))
import Control.Monad
import Data.List
import Data.Char
import Control.Monad.Trans
import Control.Monad.Fix

data Plan = Plan Code -- conditional action
  deriving (Eq, Show)

data Code = Code Expr (Maybe CompiledExpr)

instance Eq Code where
  Code e _ == Code e1 _ = e == e1

instance Show Code where
  show (Code e _) = show e

exprCode :: Expr -> Code
exprCode e = Code e Nothing

codeExpr :: Code -> Expr
codeExpr (Code e _) = e

type CompiledExpr = Eval Value

data Expr
  = ConstE Value
  | AbsE [Name] Expr
  | AppE Expr [Expr]
  | IfE Expr Expr Expr
  | LetE Name Expr Expr
  | LetNameE Name Expr Expr
  | TimeShiftE Expr Expr
  | RefE Name
  deriving (Eq, Ord)

data Value
  = IntegerV !Integer
  | NilV
  | StringV !String
  | SetV !(S.Set Value)
  | ListV ![Value]
  | BoolV !Bool
  | FunctionV !Function
  | RChV !RCh
  | SkillV !Skill
  | SkillPatternV !CharDescr !Int
  | PanelKindV !PanelKind
  | PanelV !CharDescr !Int !Int
  | PlanV !CharDescr !Int
  | SkillSlotV !CharDescr !Int
  | InvocationV !Invocation
  | ActionV !PlanAction
  | ExprV !Expr
  | AttackV !Attack
  | TimeV !Time
  | TurnV !Turn
  | InterferenceV !Interference
  | DamageV !Damage
  | ContinuationV !Continue
  | DropV !Drop
  | RandomV !Random
  deriving (Eq, Ord)

type Env = M.Map Name Entity
data Entity = Value Value | Thunk Expr
  deriving (Eq, Ord)

instance Show Entity where
  show (Value v) = show v
  show (Thunk e) = "thunk(" ++ show e ++ ")"

type Name = String
newtype Eval a = Eval { unEval :: CharDescr -> EvalContext -> Game a }

data EvalContext = EC
  { ecCurrentInv :: Maybe Invocation
  , ecCurrentAtk :: Maybe Attack
  , ecRndGen :: RndGen
  , ecNow :: Time
  }

instance Monad Eval where
  return x = Eval $ \cd ec -> return x
  Eval f >>= g = Eval $ \cd ec -> do
    r <- f cd ec
    unEval (g r) cd ec

getNow :: Eval ETime
getNow = liftM ecNow $ getContext

withNow :: ETime -> Eval a -> Eval a
withNow nn action = Eval $ \self ec -> unEval action self ec{ ecNow = nn }

getSelf :: Eval CharDescr
getSelf = Eval $ \self _ -> return self

getContext :: Eval EvalContext
getContext = Eval $ \_ ec -> return ec

liftGame :: Game a -> Eval a
liftGame x = Eval $ \_ _ -> x

unliftEval :: Eval a -> Eval (Game a)
unliftEval action = Eval $ \cd ec -> return $ unEval action cd ec

instance MonadIO Eval where
  liftIO = liftGame . liftIO

instance MonadFix Eval where
  mfix f = Eval $ \cd ec -> mfix (\x -> unEval (f x) cd ec)

newtype Random = Random { unRandom :: RndGen }

instance Eq Random where
  Random a == Random b = show a == show b

instance Ord Random where
  compare (Random a) (Random b) = compare (show a) (show b)

instance Show Random where
  show (Random a) = show a

type ETime = Time

data Function 
  = Fun { fcnName :: !Name, fcnFun :: !([Value] -> Eval Value) }
  | Fun1 { fcnName :: !Name, fcnFun1 :: !(Value -> Eval Value) }
  | Fun2 { fcnName :: !Name, fcnFun2 :: !(Value -> Value -> Eval Value) }

instance Eq Function where
  a == b = fcnName a == fcnName b

instance Ord Function where
  compare a b = compare (fcnName a) (fcnName b)

instance Show Value where
  show = render . pprValue

pprValue :: Value -> Doc
pprValue v = case v of
  IntegerV n -> text $ show n
  NilV -> text "nil"
  StringV str -> text $ showStringLit str
  SetV se -> braces $ sepByColon $ map pprValue $ S.toList se
  ListV lst -> brackets $ sepByColon $ map pprValue lst
  SkillV sk -> text $ skName sk
  SkillPatternV cd n -> text $ "#<skill-pattern " ++ show (cdToInt cd) ++ " " ++ show n ++ ">"
  PanelKindV pk -> text $ showKind pk
  RChV Opp -> text "opp"
  RChV Self -> text "self"
  BoolV True -> text "true"
  BoolV False -> text "false"
  PanelV cd m n -> text $ unwords ["#<panel", show (cdToInt cd), show m, show n ++ ">"]
  SkillSlotV cd n -> text $ "#<skill-slot " ++ show (cdToInt cd) ++ " " ++ show n ++ ">"
  InvocationV inv -> evt "invocation" inv
  FunctionV fcn -> text $ "#<function " ++ fcnName fcn ++ ">"
  ActionV pa -> cat [text "#<command [", showAction pa, text "]>"]
  ExprV expr -> cat [text "<", pprExpr expr, text ">"]
  PlanV cd n -> text $ "#<plan " ++ show (cdToInt cd) ++ " " ++ show n ++ ">"
  AttackV atk -> evt "attack" atk
  TimeV tm -> text $ "#<time " ++ show tm ++ ">"
  TurnV tm -> text $ "#<turn " ++ show (turnNumber tm) ++ ">"
  InterferenceV int -> evt "block" int
  DamageV dmg -> evt "damage" dmg
  ContinuationV cont -> evt "continuation" cont
  DropV drp -> evt "drop" drp
  RandomV gen -> text $ "#<random " ++ show gen ++ ">"
  where
    evt typ e = text $ "#<" ++ typ ++ " " ++ show (time e) ++ ">"
    cdToInt First = 0
    cdToInt Second = 1
    sepByColon xs = fcat $ intersperse (text ":") xs
    showAction pa = sepByColon $ pris ++ tgts ++ pats ++ holds
      where
        pris = map (text . spmName) $ paSkillPriority pa
        tgts = map pprTgt $ M.toList $ paSkillTarget pa
        pats = case paSkillPattern pa of
          Nothing -> []
          Just x -> [text $ "set_ss(" ++ show x ++ ")"]
        holds = map pprHold $ M.toList $ paHold pa

        pprTgt (slot, tgts) = fcat [text "target(", text $ show slot, text "|{", sepByColon (map pprT $ S.toList tgts), text "})"]
        pprT (SkillTarget cd n) = text $ "s(" ++ show cd ++ "|" ++ show n ++ ")"
        pprT (PanelTarget cd m n) = text $ "pa(" ++ show cd ++ "|" ++ show m ++ "|" ++ show  n ++ ")"
        pprT (PlanTarget cd n) = text $ "pl(" ++ show cd ++ "|" ++ show n ++ ")"
        pprHold (u, True) = text $ "hold(" ++ show u ++ ")"
        pprHold (u, False) = text $ "nohold(" ++ show u ++ ")"

showStringLit :: String -> String
showStringLit x = "\"" ++ concatMap showChar x ++ "\""
  where
    showChar c
      | isPrint c && c /= '\\' && c /= '"' = [c]
      | otherwise = "\\" ++ show (fromEnum c)

instance Show Expr where
  show = render . pprExpr

pprExpr :: Expr -> Doc
pprExpr e = case e of
  ConstE v -> pprValue v
  AbsE args body -> list [text "lambda", list (map text args), pprExpr body]
  AppE fn args -> brackets $ fsep $ pprFun fn : map pprExpr args
  IfE cond tb fb -> list [text "if", pprExpr cond, pprExpr tb, pprExpr fb]
  LetE var bi bo -> list [text "let", text var, pprExpr bi, pprExpr bo]
  LetNameE var bi bo -> list [text "letname", text var, pprExpr bi, pprExpr bo]
  TimeShiftE a b -> list [text "time-shift", pprExpr a, pprExpr b]
  RefE var -> text var
  where
    list xs = parens $ fsep xs
    pprFun (ConstE (FunctionV fcn)) = text $ "F(" ++ fcnName fcn ++ ")"
    pprFun e = pprExpr e

data RCh = Self | Opp
  deriving (Show, Eq, Ord)

data PlanAction
  = PA
  { paSkillPriority :: ![SkillPriorityModifier]
  , paSkillTarget :: !(M.Map Int (S.Set SkillTarget))
  , paSkillPattern :: !(Maybe Int)
  , paHold :: !(M.Map Invocation Bool)
  , paPlanModify :: !(M.Map Int [Function])
  }
  deriving (Eq, Ord)

data SkillPriorityModifier = 
  SPM
    String 
    (forall a. (a -> Int) -> [(a, Bool)] -> [(a, Bool)])

instance Show SkillPriorityModifier where
  show (SPM name _) = name

instance Eq SkillPriorityModifier where
  SPM name _ == SPM name1 _ = name == name1

instance Ord SkillPriorityModifier where
  compare (SPM name _) (SPM name1 _) = compare name name1

spmName :: SkillPriorityModifier -> String
spmName (SPM name _) = name

instance Monoid PlanAction where
  mappend
    PA
    { paSkillPriority = pr0
    , paSkillTarget = tg0
    , paSkillPattern = pa0
    , paHold = ho0
    , paPlanModify = pm0
    }
    PA
    { paSkillPriority = pr1
    , paSkillTarget = tg1
    , paSkillPattern = pa1
    , paHold = ho1
    , paPlanModify = pm1
    } = PA
    { paSkillPriority = pr1 ++ pr0
    , paSkillTarget = M.union tg0 tg1
    , paSkillPattern = mplus pa0 pa1
    , paHold = M.union ho0 ho1
    , paPlanModify = M.unionWith (++) pm0 pm1
    }
  mempty = PA
    { paSkillPriority = []
    , paSkillTarget = M.empty
    , paSkillPattern = Nothing
    , paHold = M.empty
    , paPlanModify = M.empty
    }

data PlanHint
  = PH
  { phSkillPriority :: !Bool
  , phSkillTarget :: !(S.Set Skill)
  , phSkillPattern :: !Bool
  , phHold :: !Bool
  , phModify :: !Bool
  }
  deriving (Show, Eq)

instance Monoid PlanHint where
  mempty = PH
    { phSkillPriority = False
    , phSkillTarget = S.empty
    , phSkillPattern = False
    , phHold = False
    , phModify = False
    }
  mappend
    PH
    { phSkillPriority = pri0
    , phSkillTarget = st0
    , phSkillPattern = sp0
    , phHold = ho0
    , phModify = mo0
    }
    PH
    { phSkillPriority = pri1
    , phSkillTarget = st1
    , phSkillPattern = sp1
    , phHold = ho1
    , phModify = mo1
    } =
    PH
    { phSkillPriority = pri0 || pri1
    , phSkillTarget = S.union st0 st1
    , phSkillPattern = sp0 || sp1
    , phHold = ho0 || ho1
    , phModify = mo0 || mo1
    }

