module Run
  ( planAction
  , askSkillPattern
  , askHoldOrNot
  , askTargets
  , askPriority
  ) where

import Game
import Plan
import ExprUtils
import Eval

import qualified Data.Set as S
import qualified Data.Map as M
import Control.Monad
import Data.Monoid
import Data.Maybe

type ActionContext = (Maybe Invocation, Maybe Attack)

nullActionContext :: ActionContext
nullActionContext = (Nothing, Nothing)

askPriority :: CharDescr -> Maybe Invocation -> Maybe Attack -> Game ((a -> Int) -> [a] -> [(a, Bool)])
askPriority cd minv matk = do
  PA{ paSkillPriority = spms } <- planActionShortcut cd phSkillPriority (minv, matk)
  let spmFun (SPM _ f) = f
  return $ \proj -> (foldr (.) id $ map (($proj) . spmFun) spms) . map (\x -> (x, False))

askTargets :: Invocation -> Game (S.Set SkillTarget)
askTargets inv = do
  PA{ paSkillTarget = st } <- planActionShortcut cd (S.member sk . phSkillTarget) nullActionContext
  return $ fromMaybe S.empty $ M.lookup (invSlot inv) st
  where
    cd = invControler inv
    sk = invSkill inv

askSkillPattern :: CharDescr -> Game (Maybe Int)
askSkillPattern cd = do
  PA{ paSkillPattern = csp } <- planActionShortcut cd phSkillPattern nullActionContext
  return csp

askHoldOrNot :: CharDescr -> Game (Invocation -> Bool)
askHoldOrNot cd = do
  hss <- getHSs cd
  if not $ any hsEffective hss then return (const $ error "askHoldOrNot:inpossible") else do
    PA{ paHold = m } <- planActionShortcut cd phHold nullActionContext
    return $ \inv -> fromMaybe False $ M.lookup inv m

planAction :: CharDescr -> Game PlanAction
planAction cd = planActionShortcut cd (const True) nullActionContext

planActionShortcut :: CharDescr -> (PlanHint -> Bool) -> ActionContext -> Game PlanAction
planActionShortcut cd prop (minv, matk) = do
  now <- tickTime
  plans <- getPresentPlans cd
  process now mempty M.empty plans
  where
    process :: Time -> PlanAction -> M.Map Int Expr -> [PlanState] -> Game PlanAction
    process now a c ps | a `seq` c `seq` ps `seq` False = undefined
    process now action changes [] = return action
    process now action changes (ps@PS{ psHint = hint, psBound = bound }:rest) = do
      gen <- newRndGen
      if bound == 0 && (prop hint || phModify hint) || M.member idx changes
        then do
          let ec = EC{ ecCurrentInv = minv, ecCurrentAtk = matk, ecRndGen = gen, ecNow = now }
          r <- case M.lookup idx changes of
            Just exp -> runCode cd ec $ exprCode exp
            Nothing -> runCode cd ec $ planCode plan
          changes' <- addChanges ec (paPlanModify r) changes idx
          process now (mappend action r) changes' rest
        else process now action changes rest
      where
        idx = psIndex ps
        plan = psPlan ps

        addChanges ec pm changes idx = foldM f changes $ M.toList pm
          where
            f chs (n, funs)
              | n <= idx = return chs
              | otherwise = case M.lookup n chs of
                Just mod -> use mod
                Nothing -> do
                  Just ps <- getPlanState cd n
                  use $ planExpr $ psPlan ps
              where
                use e = do
                  e' <- foldM trans e funs
                  return $ M.insert n e' chs
            trans expr fun = do
              r <- runIgn cd ec $ applyFun1 fun (ExprV expr)
              case r of
                Nothing -> return expr
                Just (ExprV e') -> return e'
                Just v -> do
                  mes $ "error: plan modifier returned a non-function: " ++ show v
                  return expr

runCode :: CharDescr -> EvalContext -> Code -> Game PlanAction
runCode cd ec (Code exp mc) = do
  r <- runIgn cd ec run
  case r of
    Just (ActionV v) -> return v
    _ -> return mempty
  where
    run = case mc of
      Just e -> e
      Nothing -> eval exp

runIgn :: CharDescr -> EvalContext -> Eval a -> Game (Maybe a)
runIgn cd ec action = do
  r <- runEvalWithContext cd action ec
  case r of
    Right v -> return $ Just v
    Left msg -> do
      mes $ "execution error: " ++ msg
      return Nothing
